"""Parser implementations for (mostly) builtin types."""

from __future__ import annotations

import contextlib
import inspect
import string
import typing

import attrs
import typing_extensions

from disnake_compass.impl.parser import base as parser_base

if typing.TYPE_CHECKING:
    from disnake_compass.api import parser as parser_api

__all__: typing.Sequence[str] = (
    "BoolParser",
    "CollectionParser",
    "FloatParser",
    "IntParser",
    "StringParser",
    "TupleParser",
    "UnionParser",
)

_NoneType: type[None] = type(None)
_NONES = (None, _NoneType)
_INT_CHARS = string.digits + string.ascii_lowercase

_CollectionT = typing_extensions.TypeVar(  # Simplest iterable container object.
    "_CollectionT",
    bound=typing.Collection[object],
    default=typing.Collection[str],
)
_TupleT = typing_extensions.TypeVar("_TupleT", bound=tuple[typing.Any, ...])
_T = typing_extensions.TypeVar("_T")

# NONE


@parser_base.register_parser_for(_NoneType)
@attrs.define(slots=True)
class NoneParser(parser_base.Parser[None]):
    r"""Parser implementation for :obj:`None`.

    Mainly relevant for :obj:`~typing.Optional`\[...] parsers.

    Parameters
    ----------
    strict: bool
        Whether this parser should be set to :attr:`strict` mode.

        To prevent unforeseen bugs, this defaults to :obj:`True`.

    """

    strict: bool = attrs.field(default=True, kw_only=True)
    """Whether this parser is set to strict mode.

    See :meth:`loads` and :meth:`dumps` for the implications of strict-mode.
    """

    async def loads(self, argument: str, /) -> None:
        """Load ``None`` from a string.

        If :attr:`strict` is set to ``True``, this will fail if the
        provided ``argument`` isn't the empty string (``""``). Otherwise,
        this parser will *always* return ``None``.

        Parameters
        ----------
        argument:
            The string that is to be converted to ``None``.

        Raises
        ------
        ValueError:
            The parser is in strict mode, and the provided ``argument`` was not
            the empty string.

        """
        if not argument or not self.strict:
            return None  # noqa: RET501

        msg = f"Strict `NoneParser`s can only load the empty string, got {argument!r}."
        raise ValueError(msg)

    async def dumps(self, argument: None, /) -> str:
        """Dump ``None`` into a string.

        If :attr:`strict` is set to ``True``, this will fail if the
        provided ``argument`` isn't exactly ``None``. Otherwise,
        this parser will *always* return ``None``.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        Raises
        ------
        ValueError:
            The parser is in strict mode, and the provided ``argument`` was not
            ``None``.

        """
        if argument is None or not self.strict:
            return ""

        msg = f"Strict `NoneParser`s can only dump `None`, got {argument!r}."
        raise ValueError(msg)


# INT / FLOAT


def _removesuffix(string: str, suffix: str) -> str:
    return string[: -len(suffix) if string.endswith(suffix) else None]


def dumps_float(number: float) -> str:
    """Dump a float to a string.

    Ensures trailing .0 are stripped to minimise float character length.
    """
    return _removesuffix(str(number), ".0")


# TODO: decimal.Decimal parser for decimal numbers with set precision?


@parser_base.register_parser_for(float)
@attrs.define(slots=True)
class FloatParser(parser_base.Parser[float]):
    r"""Parser implementation for :class:`float`\s."""

    async def loads(self, argument: str, /) -> float:
        """Load a floating point number from a string.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a floating point number.

        """
        return float(argument)

    async def dumps(self, argument: float, /) -> str:
        """Dump a floating point number into a string.

        Strips trailing ".0" where possible.

        Parameters
        ----------
        argument:
            The string that is to be converted into a floating point number.

        """
        return dumps_float(argument)


def _validate_base(_instance: object, _attribute: object, base: int) -> None:
    if not 2 <= base <= 36:  # noqa: PLR2004
        msg = "Base must be between 2 and 36."
        raise ValueError(msg)


@parser_base.register_parser_for(int)
@attrs.define(slots=True)
class IntParser(parser_base.Parser[int]):
    r"""Parser implementation for :class:`int`\s.

    Parameters
    ----------
    signed:
        Whether the parser supports signed integers.
        Defaults to ``True``.
    base:
        The base to use to use for storing integers.
        This is limited to ``2 <= base <= 36``.
        Defaults to ``36``.

    """

    signed: bool = attrs.field(default=True)
    """Whether the parser supports signed integers."""
    base: int = attrs.field(default=36, validator=_validate_base)
    """The base to use to use for storing integers.
    This is limited to ``2 <= base <= 36`` as this is the range supported by
    python's :class:`int` constructor.

    If a greater base is required, a custom integer parser will have to be
    implemented.
    """

    async def loads(self, argument: str, /) -> int:
        r"""Load an integer from a string.

        Parameters
        ----------
        argument:
            The string that is to be converted to an ``int``.

        Raises
        ------
        ValueError:
            The parser has :attr:`signed` set to ``False`` but the argument
            was a negative number.
            Alternatively, the provided argument is not a valid integer at all.

        """
        result = int(argument, self.base)
        if not self.signed and result < 0:
            msg = "Unsigned numbers cannot be < 0."
            raise ValueError(msg)

        return result

    async def dumps(self, argument: int, /) -> str:
        """Dump an integer into a string.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        # Try to short-circuit as much as possible
        if argument < self.base:
            return str(argument)
        if self.base == 2:  # noqa: PLR2004
            return f"{argument:b}"
        if self.base == 8:  # noqa: PLR2004
            return f"{argument:o}"
        if self.base == 10:  # noqa: PLR2004
            return str(argument)
        if self.base == 16:  # noqa: PLR2004
            return f"{argument:x}"

        # Can't short-circuit, convert to string manually.
        digits: list[int] = []
        while argument:
            digits.append(argument % self.base)
            argument //= self.base

        return "".join(_INT_CHARS[d] for d in reversed(digits))


# BOOL

_DEFAULT_TRUES = frozenset(["true", "t", "yes", "y", "1"])
_DEFAULT_FALSES = frozenset(["false", "f", "no", "n", "0"])


@parser_base.register_parser_for(bool)
@attrs.define(slots=True)
class BoolParser(parser_base.Parser[bool]):
    """Parser type with support for bools.

    This parser type can be supplied with a collection of strings for the
    values that should be considered true and false. By default,
    ``"true", "t", "yes", "y", "1"`` are considered ``True``, while
    ``"false", "f", "no", "n", "0"`` are considered ``False``. Note that this
    is case-insensitive.

    Parameters
    ----------
    trues:
        Values that should be considered ``True`` by this parser.
    falses:
        Values that should be considered ``False`` by this parser.

    """

    trues: typing.Collection[str] = attrs.field(factory=_DEFAULT_TRUES.copy)
    """A collection of values that should be considered ``True`` by this parser."""
    falses: typing.Collection[str] = attrs.field(factory=_DEFAULT_FALSES.copy)
    """A collection of values that should be considered ``False`` by this parser."""

    async def loads(self, argument: str, /) -> bool:
        """Load a boolean from a string.

        Parameters
        ----------
        argument:
            The string that is to be converted into a boolean.

        Raises
        ------
        ValueError:
            The string is not in :attr:`trues` or :attr:`falses`.

        """
        if argument in self.trues:
            return True
        if argument in self.falses:
            return False

        trues_str = ", ".join(map(repr, self.trues))
        falses_str = ", ".join(map(repr, self.falses))
        msg = (
            f"Failed to parse {argument!r} into a bool. Expected any of"
            f" {trues_str} for True, or any of {falses_str} for False."
        )
        raise ValueError(msg)

    async def dumps(self, argument: bool, /) -> str:  # noqa: FBT001
        """Dump a boolean into a string.

        By default, this opts to dump as ``"1"`` for ``True`` or ``"0"`` for
        ``False`` to only use one character's worth of space from the custom id.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return "1" if argument else "0"


# STRING


@parser_base.register_parser_for(str)
@attrs.define(slots=True)
class StringParser(parser_base.Parser[str]):
    """Parser type with support for strings.

    Both loads and dumps are essentially no-ops.
    """

    async def loads(self, argument: str, /) -> str:
        """Load a string from a string.

        Parameters
        ----------
        argument:
            The string that is to be converted into a string.

        """
        return argument

    async def dumps(self, argument: str, /) -> str:
        """Dump a string into a string.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return argument


def _resolve_collection(type_: type[_CollectionT]) -> type[_CollectionT]:
    # ContainerParser itself does not support tuples.
    if issubclass(type_, tuple):
        msg = (
            f"{CollectionParser.__name__}s do not support tuples. Please use a "
            f"{TupleParser.__name__} instead."
        )
        raise TypeError(msg)

    if not typing_extensions.is_protocol(type_) and not inspect.isabstract(type_):
        # Concrete type, return as-is.
        return type_

    # Try to resolve an abstract type to a valid concrete structural subtype.
    if issubclass(type_, typing.Sequence):
        return typing.cast(type[_CollectionT], list)

    if issubclass(type_, typing.AbstractSet):
        return typing.cast(type[_CollectionT], set)

    msg = f"Cannot infer a concrete type for abstract type {type_.__name__!r}."
    raise TypeError(msg)


# NOTE: TupleParser *must* be registered before CollectionParser!


@parser_base.register_parser_for(tuple, priority=10)
@attrs.define(slots=True, init=False)
class TupleParser(parser_base.Parser[_TupleT]):
    r"""Parser type with support for :class:`tuple`\s.

    The benefit of a tuple parser is fixed-length checks and the ability to set
    multiple types. For example, a ``Tuple[str, int, bool]`` parser will
    actually return a tuple with a ``str``, ``int``, and ``bool`` inside.

    Parameters
    ----------
    *inner_parsers:
        The parsers to use to parse the items inside the tuple.

        Defaults to a single string parser, i.e. a one-element tuple containing
        exactly one string.
    sep:
        The separator to use.

        Defaults to ",".

    """

    inner_parsers: tuple[parser_api.Parser[typing.Any], ...]
    """The parsers to use to parse the items inside the tuple.

    These define the inner types and the allowed number of items in the in the
    tuple.
    """
    sep: str
    """The separator to use.

    Can be any string, though a single character is recommended.

    .. warning::
        Ensure that this does **not** match
        :attr:`ComponentManager.sep <disnake_compass.impl.manager.ComponentManager.sep>`
        on the component manager that corresponds to this parser's component.
        """
    tuple_cls: type[_TupleT]
    """The tuple type to use.

    This mainly exists to support NamedTuples.
    """
    # NOTE: NamedTuple does not show up in the class MRO so get_parser cannot
    #       tell the difference. NamedTuple initialisation is different from
    #       normal tuple initialisation, so we need to special-case it.

    def __init__(
        self,
        *inner_parsers: parser_api.Parser[typing.Any],
        sep: str = ",",
        tuple_cls: type[_TupleT] = tuple,
    ) -> None:
        self.inner_parsers = inner_parsers or (StringParser.default(str),)
        self.sep = sep
        self.tuple_cls = tuple_cls

    @classmethod
    def default(cls, type_: type[_TupleT], /) -> typing_extensions.Self:  # noqa: D102
        if hasattr(type_, "__annotations__"):
            # This is a namedtuple.
            args = typing.get_type_hints(type_).values()
        else:
            args = typing.get_args(type_)

        inner_parsers = [parser_base.get_parser(arg) for arg in args]
        return cls(*inner_parsers, tuple_cls=type_)

    async def loads(self, argument: str, /) -> _TupleT:
        """Load a tuple from a string.

        Parameters
        ----------
        argument:
            The string that is to be converted into a tuple.

            This is split over :attr:`sep` and then each individual substring
            is passed to its respective inner parser.

        Raises
        ------
        RuntimeError:
            The number of substrings after splitting does not match the number
            of inner parsers.

        """
        parts = argument.split(self.sep)

        # NamedTuples should be instantiated using _make.
        initialiser = getattr(self.tuple_cls, "_make", self.tuple_cls)
        return initialiser(
            [
                await parser.loads(part)
                for parser, part in zip(self.inner_parsers, parts, strict=True)
            ],
        )

    async def dumps(self, argument: _TupleT, /) -> str:
        """Dump a tuple into a string.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        Raises
        ------
        :class:`RuntimeError`:
            The length of the ``argument`` tuple does not match the number of
            inner parsers.

        """
        return self.sep.join(
            [
                await parser.dumps(part)
                for parser, part in zip(self.inner_parsers, argument, strict=True)
            ],
        )


@parser_base.register_parser_for(typing.Collection)
@attrs.define(slots=True, init=False)
class CollectionParser(parser_base.Parser[_CollectionT]):
    r"""Parser type with support for :class:`typing.Collection`\s.

    This supports types such as :class:`list`, :class:`set`, etc.; but also
    abstract types such as :class:`typing.Collection` itself.

    .. note:: This parser does **not** support :class:`tuple`\s.

    Parameters
    ----------
    inner_parser:
        The parser to use to parse the items inside the collection.

        This defines the inner type for the collection.

        Defaults to a :class:`StringParser`.
    collection_type:
        The type of collection to use. This does not specify the inner type.

        Defaults to :class:`list`.
    sep:
        The separator to use.

        Can be any string, though a single character is recommended.

        Defaults to ``","``.

    """

    inner_parser: parser_api.Parser[typing.Any]
    """The parser to use to parse the items inside the collection.

    Note that, unlike a :class:`TupleParser`, a collection parser requires all
    items to be of the same type.
    """
    collection_type: type[_CollectionT]
    """The collection type for this parser.

    This is the type that holds the items, e.g. a :class:`list`.

    .. note::
        This has to support being instantiated from an iterable as
        ``collection_type(iterable_of_values)``; e.g. ``set([1,2,3])``
    """
    sep: str
    """The separator to use.

    Can be any string, though a single character is recommended.

    .. warning::
        Ensure that this does **not** match
        :attr:`ComponentManager.sep <disnake_compass.impl.manager.ComponentManager.sep>`
        on the component manager that corresponds to this parser's component.
        """

    def __init__(
        self,
        inner_parser: parser_api.Parser[typing.Any] | None = None,
        *,
        collection_type: type[_CollectionT] | None = None,
        sep: str = ",",
    ) -> None:
        self.sep = sep
        self.collection_type = typing.cast(  # Pyright do be whack sometimes.
            type[_CollectionT],
            list if collection_type is None else _resolve_collection(collection_type),
        )
        self.inner_parser = StringParser.default(str) if inner_parser is None else inner_parser

    @classmethod
    def default(  # noqa: D102
        cls,
        type_: type[_CollectionT],
        /,
    ) -> typing_extensions.Self:
        origin = typing.get_origin(type_)
        args = typing.get_args(type_)

        inner_type = next(iter(args), str)  # Get first element, default to str
        inner_parser = parser_base.get_parser(inner_type)
        return cls(inner_parser, collection_type=origin)

    async def loads(self, argument: str, /) -> _CollectionT:
        """Load a collection from a string.

        Parameters
        ----------
        argument:
            The string that is to be converted into a collection.

            This is split over :attr:`sep` and then each individual substring
            is passed to its respective inner parser.

        """
        # TODO: Maybe make this a generator instead of a list?
        parsed = [
            await self.inner_parser.loads(part)
            for part in argument.split(self.sep)
            if not part.isspace()  # TODO: Verify if this should be removed
        ]

        return self.collection_type(parsed)  # pyright: ignore[reportCallIssue]

    async def dumps(self, argument: _CollectionT, /) -> str:
        """Dump a collection into a string.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return ",".join([await self.inner_parser.dumps(part) for part in argument])


@parser_base.register_parser_for(typing.Union)  # pyright: ignore[reportArgumentType]
@attrs.define(slots=True, init=False)
class UnionParser(parser_base.Parser[_T], typing.Generic[_T]):
    r"""Parser type with support for :class:`~typing.Union`\s.

    The provided parsers are sequentially tried until one passes. If none work,
    an exception is raised instead.

    .. important::
        Unlike :class:`NoneParser`, :attr:`strict` for this class is set
        via the underlying none parser, if any. Note that this class *does*
        proxy it through the :attr:`strict` property, which supports both
        getting and setting; but **only** if this parser is :attr:`optional`.

    Parameters
    ----------
    *inner_parsers:
        The parsers with which to sequentially try to parse the argument.

        ``None`` can be provided as one of the parameters to make it optional;
        this will automatically add a **strict** :class:`NoneParser`.

    """

    inner_parsers: typing.Sequence[parser_api.Parser[typing.Any]]
    """The parsers with which to sequentially try to parse the argument."""
    optional: bool
    """Whether this parser is optional."""

    def __init__(
        self,
        *inner_parsers: parser_api.Parser[typing.Any] | None,
    ) -> None:
        if len(inner_parsers) < 2:  # noqa: PLR2004
            msg = "A Union requires two or more type arguments."
            raise TypeError(msg)

        self.optional = False
        self.inner_parsers = []
        for parser in inner_parsers:
            if parser in _NONES:
                self.optional = True

            else:
                self.inner_parsers.append(parser)

        # Ensure the NoneParser always comes last.
        if self.optional:
            self.inner_parsers.append(NoneParser.default(_NoneType))

    @classmethod
    def default(cls, type_: type[_T], /) -> typing_extensions.Self:  # noqa: D102
        args = typing.get_args(type_)

        inner_parsers = [parser_base.get_parser(arg) for arg in args]
        return cls(*inner_parsers)

    @property
    def strict(self) -> bool:
        """Whether this parser is strict.

        This is only meaningful when the parser is optional, in which case it
        directly reflects :attr:`NoneParser.strict`.
        """
        if not self.optional:
            return True

        none_parser = self.inner_parsers[-1]
        assert isinstance(none_parser, NoneParser)
        return none_parser.strict

    @strict.setter
    def strict(self, strict: bool) -> None:
        if not self.optional:
            cls = type(self).__name__
            msg = f"{cls}.strict can only be set when {cls}.optional is True."
            raise RuntimeError(msg)

        none_parser = self.inner_parsers[-1]
        assert isinstance(none_parser, NoneParser)
        none_parser.strict = strict

    async def loads(self, argument: str, /) -> _T:
        """Load a union of types from a string.

        If :attr:`optional` is ``True`` and :attr:`strict` is ``False``, this
        returns ``None`` if all parsers fail. Otherwise, an exception is raised.

        Parameters
        ----------
        argument:
            The string that is to be converted into one of the types in this union.

            Each inner parser is tried in the order they are provided, and the
            first to load the ``argument`` successfully short-circuits and
            returns.

        Raises
        ------
        :class:`RuntimeError`
            None of the :attr:`inner_parsers` succeeded to load the ``argument``.

        """
        if not argument and self.optional:
            # Quick-return: if no argument was provided and the parser is
            # optional, just return None without trying any parsers.
            return typing.cast(_T, None)

        # Try all parsers sequentially. If any succeeds, return the result.
        for parser in self.inner_parsers:
            with contextlib.suppress(Exception):
                return await parser.loads(argument)

        msg = "Failed to parse input to any type in the Union."
        raise RuntimeError(msg)

    async def dumps(self, argument: _T, /) -> str:
        """Dump a union of types into a string.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

            This finds a parser that is registered for the type of the provided
            ``argument``.

        Raises
        ------
        :class:`RuntimeError`:
            None of the :attr:`inner_parsers` succeeded to dump the ``argument``.

        """
        if not argument and self.optional:
            return ""

        # TODO: Maybe add contextlib.suppress?
        for parser in self.inner_parsers:
            with contextlib.suppress(Exception):
                return await parser.dumps(argument)

        if self.optional and not self.strict:
            # Act like the NoneParser with strict=False dumped the argument.
            return ""

        msg = f"Failed to parse input {argument!r} to any type in the Union."
        raise RuntimeError(msg)


@parser_base.register_parser_for(typing.Literal)  # pyright: ignore[reportArgumentType]
@attrs.define(slots=True, init=False)
class LiteralParser(parser_base.Parser[_T], typing.Generic[_T]):
    options: typing.Sequence[_T]
    inner_parser: parser_api.Parser[_T]

    def __init__(
        self,
        *options: _T,
        inner_parser: parser_api.Parser[_T],
    ) -> None:
        self.options = options
        self.inner_parser = inner_parser

    @classmethod
    def default(cls, type_: type[_T], /) -> typing_extensions.Self:
        assert typing.get_origin(type_) == typing.Literal
        args: tuple[_T] = typing.get_args(type_)

        args_iter = iter(args)
        arg_type = type(next(args_iter))
        for arg in args_iter:
            if not isinstance(arg, arg_type):
                msg = (
                    "Literal parsers only support literals where all options are of the same type."
                )
                raise TypeError(msg)

        return cls(*args, inner_parser=parser_base.get_parser(arg_type))

    async def loads(self, argument: str, /) -> _T:
        value = await self.inner_parser.loads(argument)

        assert value in self.options

        return value

    async def dumps(self, argument: _T, /) -> str:
        if argument not in self.options:
            msg = (
                f"{argument!r} is not a valid option for this parser."
                f" Expected any of {', '.join(map(repr, self.options))}"
            )
            raise ValueError(msg)

        return await self.inner_parser.dumps(argument)



"""Small utilities for Sphinx documentation."""

import importlib.util
import inspect
import os
import sys
import typing

import sphinx.config
import sphinx_autodoc_typehints as original
from sphinx.util import inspect as sphinx_inspect

__all__ = (
    "apply_autodoc_typehints_patch",
    "format_annotation",
    "get_module_path",
    "make_linkcode_resolver",
)


NoneType = type(None)


def get_module_path() -> str:
    """Get the filepath for the module."""
    _spec = importlib.util.find_spec("disnake_compass")
    if not (_spec and _spec.origin):
        raise RuntimeError  # this should never happen

    return os.path.dirname(_spec.origin)


def make_linkcode_resolver(
    module_path: str,
    repo_url: str,
    git_ref: str,
) -> typing.Callable[[str, typing.Mapping[str, typing.Any]], str | None]:
    """Return a linkcode resolver for the provided module path and repo data."""

    def linkcode_resolve(
        domain: str,
        info: typing.Mapping[str, typing.Any],
    ) -> str | None:
        if domain != "py":
            return None

        try:
            obj: typing.Any = sys.modules[info["module"]]
            for part in info["fullname"].split("."):
                obj = getattr(obj, part)
            obj = inspect.unwrap(obj)

            if isinstance(obj, property):
                obj = inspect.unwrap(obj.fget)  # pyright: ignore

            path: typing.Any = inspect.getsourcefile(obj)
            path = os.path.relpath(path, start=module_path)
            src, lineno = inspect.getsourcelines(obj)

        except Exception:  # noqa: BLE001
            return None

        path = f"{path}#L{lineno}-L{lineno + len(src) - 1}"
        return f"{repo_url}/blob/{git_ref}/src/disnake_compass/{path}"

    return linkcode_resolve


def format_annotation(  # noqa: PLR0911, PLR0912, PLR0915
    annotation: typing.Any,  # noqa: ANN401
    config: sphinx.config.Config,
) -> str:
    """Format the annotation."""
    if typehints_formatter := getattr(config, "typehints_formatter", None):
        formatted = typehints_formatter(annotation, config)
        if formatted is not None:
            return formatted

    # Special cases
    if isinstance(annotation, typing.ForwardRef):
        return annotation.__forward_arg__
    if annotation in (None, NoneType):
        return ":py:obj:`None`"
    if annotation is Ellipsis:
        return ":py:data:`...<Ellipsis>`"

    if isinstance(annotation, tuple):
        return original.format_internal_tuple(annotation, config)  # pyright: ignore

    if isinstance(annotation, sphinx_inspect.TypeAliasForwardRef):
        return str(annotation)

    try:
        module = original.get_annotation_module(annotation)
        class_name = original.get_annotation_class_name(annotation, module)
        args = original.get_annotation_args(annotation, module, class_name)
    except ValueError:
        return str(annotation).strip("'")

    # Redirect all typing_extensions types to the stdlib typing module
    if module == "typing_extensions":
        module = "typing"

    elif module == "_io":
        module = "io"

    full_name = f"{module}.{class_name}" if module != "builtins" else class_name
    fully_qualified: bool = getattr(config, "typehints_fully_qualified", False)
    prefix = "" if fully_qualified or full_name == class_name else "~"
    role = (
        "data"
        if module == "typing" and class_name in original._PYDATA_ANNOTATIONS  # pyright: ignore
        else "class"
    )
    args_format = "\\[{}]"
    formatted_args: str = ""

    always_use_bars_union: bool = getattr(config, "always_use_bars_union", True)
    is_bars_union = full_name == "types.UnionType" or (
        always_use_bars_union and type(annotation).__qualname__ == "_UnionGenericAlias"
    )
    if is_bars_union:
        full_name = ""

    # Some types require special handling
    if full_name == "typing.NewType":
        args_format = f"\\(``{annotation.__name__}``, {{}})"
        role = "class" if sys.version_info >= (3, 10) else "func"

    # NOTE: Modified to return :py:data:`NameOfTypeVar`
    elif full_name in {"typing.TypeVar", "typing.ParamSpec"}:
        role = "data"
        full_name = annotation.__name__

    elif full_name == "typing.Optional":
        args = tuple(x for x in args if x is not NoneType)

    elif full_name in ("typing.Union", "types.UnionType") and NoneType in args:
        if len(args) == 2:
            full_name = "typing.Optional"
            role = "data"
            args = tuple(x for x in args if x is not type(None))
        else:
            simplify_optional_unions: bool = getattr(
                config,
                "simplify_optional_unions",
                True,
            )
            if not simplify_optional_unions:
                full_name = "typing.Optional"
                role = "data"
                args_format = f"\\[:py:data:`{prefix}typing.Union`\\[{{}}]]"
                args = tuple(x for x in args if x is not type(None))

    elif (
        full_name in ("typing.Callable", "collections.abc.Callable") and args and args[0] is not ...
    ):
        fmt = [format_annotation(arg, config) for arg in args]
        formatted_args = f"\\[\\[{', '.join(fmt[:-1])}], {fmt[-1]}]"

    elif full_name == "typing.Literal":
        formatted_args = f"\\[{', '.join(f'``{arg!r}``' for arg in args)}]"

    elif is_bars_union:
        return " | ".join([format_annotation(arg, config) for arg in args])

    # TODO: Maybe figure out some way of having Self actually link to the class.
    #       Low priority though.

    if args and not formatted_args:
        try:
            iter(args)
        except TypeError:
            fmt = [format_annotation(args, config)]
        else:
            fmt = [format_annotation(arg, config) for arg in args]
        formatted_args = args_format.format(", ".join(fmt))

    escape = "\\ " if formatted_args else ""
    return f":py:{role}:`{prefix}{full_name}`{escape}{formatted_args}"


def apply_autodoc_typehints_patch() -> None:
    """Apply the sphinx-ext-autodoc monkeypatch."""
    original.format_annotation = format_annotation



"""A simple CLI command to run example files with a bot token from env."""

import argparse
import importlib.util
import os
import subprocess
import sys
import typing

import dotenv

_KEY: typing.Final[str] = "EXAMPLE_TOKEN"


def _main() -> typing.NoReturn:
    # First, get the token from dotenv.
    dotenv.load_dotenv(override=True)
    token = os.getenv(_KEY)
    if not token:
        print("Token not found! Please set 'EXAMPLE_TOKEN' in .env.")
        sys.exit(1)

    choices = [
        file[:-3] for file in os.listdir(os.path.realpath("examples")) if file.endswith(".py")
    ]

    # Get the passed example using argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("example", type=str, choices=choices)

    result = parser.parse_args()
    example: str = f"examples.{result.example}"

    # Check if example module exists:
    if not importlib.util.find_spec(example):
        print(f"Cound not find example {example!r}.")
        sys.exit(1)

    # Run example with unbuffered (-u) python.
    args = [sys.executable, "-u", "-m", example]

    # Open a subprocess that runs our args list.
    print(f"Running example {example!r}...")

    try:
        with subprocess.Popen(
            args,
            # Pipe all output to stdout...
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            # Ensure proper cwd...
            cwd=os.getcwd(),
            # Set buffer mode to line-buffered...
            bufsize=1,
            # Set stdout mode to text...
            text=True,
            # Provide bot token environment variable...
            env={**os.environ, _KEY: token},
        ) as process:
            if not process.stdout:
                print("Failed to open stdout.")
                sys.exit(1)

            for line in process.stdout:
                print(line, end="")

    except KeyboardInterrupt:
        print("Received KeyboardInterrupt, stopping example.")
        sys.exit(0)

    # And finally, propagate the return code.
    sys.exit(process.returncode)


if __name__ == "__main__":
    _main()



# pyright: reportImportCycles = false
# pyright: reportWildcardImportFromLibrary = false
# ^ This is a false positive as it is confused with site-packages' disnake.

"""Utilities mostly intended for internal use.

These may also be of help when creating custom implementations of protocols
detailed in ``disnake_compass.api``.
"""



"""Protocols for components and component managers."""

from __future__ import annotations

import typing

import disnake

if typing.TYPE_CHECKING:
    import typing_extensions

__all__: typing.Sequence[str] = (
    "ComponentManager",
    "RichButton",
    "RichComponent",
    "RichSelect",
)


_T = typing.TypeVar("_T")

AnyEmoji: typing.TypeAlias = str | disnake.PartialEmoji | disnake.Emoji
MaybeCoroutine: typing.TypeAlias = _T | typing.Coroutine[None, None, _T]

ComponentT = typing.TypeVar("ComponentT", bound="RichComponent")
"""A type hint for a (subclass of) a disnake-compass component.

In practice, this will be any implementation of the :class:`RichButton`,
or :class:`RichSelect` protocols.
"""


@typing.runtime_checkable
class RichComponent(typing.Protocol):
    """The baseline protocol for any kind of component.

    Any and all components must implement this protocol in order to be properly
    handled by disnake-compass.
    """

    __slots__: typing.Sequence[str] = ()

    @classmethod
    def get_manager(cls) -> ComponentManager:
        """Get the manager that was responsible for parsing this component instance."""
        ...

    @classmethod
    def set_manager(cls, manager: ComponentManager | None, /) -> None:
        """Set the manager that was responsible for parsing this component instance."""
        ...

    @classmethod
    def get_factory(cls) -> ComponentFactory[typing_extensions.Self]:
        """Get the factory that built this component instance."""
        ...

    @classmethod
    def set_factory(cls, factory: ComponentFactory[typing_extensions.Self]) -> None:
        """Set the factory that built this component instance."""
        ...

    async def callback(self, interaction: disnake.Interaction[disnake.Client], /) -> None:
        """Run the component callback.

        This should be implemented by the user in each concrete component type.

        Parameters
        ----------
        interaction:
            The interaction that caused this button to fire.

        """
        ...

    async def as_ui_component(
        self, manager: ComponentManager | None = None, /
    ) -> disnake.ui.WrappedComponent:
        """Convert this component into a component that can be sent by disnake.

        Returns
        -------
        :class:`WrappedComponent <disnake.ui.WrappedComponent>`
            A component that can be sent by disnake, maintaining the parameters
            and custom id set on this rich component.

        """
        ...


@typing.runtime_checkable
class RichButton(RichComponent, typing.Protocol):
    """Baseline protocol for Button-like components."""

    __slots__: typing.Sequence[str] = ()

    label: str | None
    """The label of the component.

    Either or both this field or the ``emoji`` field must be set.
    """
    style: disnake.ButtonStyle
    """The style of the component.

    This dictates how the button is displayed on discord.
    """
    emoji: AnyEmoji | None
    """The emoji that is to be displayed on this button.

    Either or both this field or the ``emoji`` field must be set.
    """
    disabled: bool
    """Whether or not this button is disabled.

    A disabled button is greyed out on discord, and cannot be pressed.
    Disabled buttons can therefore not cause any interactions, either.
    """
    id: int

    async def as_ui_component(  # noqa: D102
        self, manager: ComponentManager | None = None, /
    ) -> disnake.ui.Button[None]:
        # <<Docstring inherited from RichComponent>>
        ...


@typing.runtime_checkable
class RichSelect(RichComponent, typing.Protocol):
    """Baseline protocol for Select-like components."""

    __slots__: typing.Sequence[str] = ()

    placeholder: str | None
    """The placeholder of the component.

    This shows when nothing is selected, or shows nothing if set to
    :data:`None`.
    """
    min_values: int
    """The minimum number of values the user must select.

    This must lie between 1 and 25, inclusive.
    """
    max_values: int
    """The maximum number of values the user may select.

    This must lie between 1 and 25, inclusive.
    """
    disabled: bool
    """Whether or not this button is disabled.

    A disabled select is greyed out on discord, and cannot be used.
    Disabled selects can therefore not cause any interactions, either.
    """
    id: int

    async def as_ui_component(  # noqa: D102
        self, manager: ComponentManager | None = None, /
    ) -> disnake.ui.BaseSelect[typing.Any, typing.Any, None]:
        # <<Docstring inherited from RichComponent>>
        ...


@typing.runtime_checkable
class ComponentManager(typing.Protocol):
    """The baseline protocol for component managers.

    Component managers keep track of disnake-compass' special components
    and ensure they smoothly communicate with disnake's bots. Since this relies
    on listener functionality, component managers are incompatible with
    :class:`disnake.Client`-classes.

    To register a component to a component manager, use
    :meth:`register_component`. Without registering your components, they will
    remain unresponsive.
    """

    __slots__: typing.Sequence[str] = ()

    @property
    def name(self) -> str:
        """The name of this manager.

        Used in :func:`get_manager`. This functions similar to loggers, where a
        parent-child relationship is denoted with a ".". For example, a manager
        "foo.bar" has parent "foo", which has the root manager as parent.
        """
        ...

    @property
    def children(self) -> typing.Collection[ComponentManager]:
        """The children of this component manager."""
        ...

    @property
    def components(self) -> typing.Mapping[str, type[RichComponent]]:
        """The components registered to this manager or any of its children.

        In case a custom implementation is made, special care must be taken to
        ensure that these do not desync when a child's components are updated.
        """
        ...

    @property
    def parent(self) -> ComponentManager | None:
        """The parent of this manager.

        Returns :data:`None` in case this is the root manager.
        """
        ...

    def make_identifier(self, component_type: type[RichComponent], /) -> str:
        """Make an identifier for the provided component class.

        This is used to store the component in :attr:`components`, and to
        determine which component's callback should be fired when an interaction
        is received.

        Parameters
        ----------
        component_type
            The type of component for which to make an identifier.

        Returns
        -------
        str
            The component type's identifier.

        """
        ...

    def get_identifier(self, custom_id: str, /) -> tuple[str, typing.Sequence[str]]:
        """Extract the identifier and parameters from a custom id.

        This is used to check whether the identifier is registered in
        :attr:`components`.

        Parameters
        ----------
        custom_id
            The custom id from which to extract the identifier.

        """
        ...

    def lookup_identifier(self, component_type: type[RichComponent], /) -> str:
        """Look up the identifier of an already registered component type.

        Parameters
        ----------
        component_type
            The component type for which to look up the registered identifier.

        """
        ...

    async def make_custom_id(self, component: RichComponent, /) -> str:
        """Make a custom id from the provided component.

        This can then be used later to reconstruct the component without any
        state or data loss.

        Parameters
        ----------
        component
            The component for which to create a custom id.

        Returns
        -------
        str
            A custom id that fully represents the provided component.

        """
        ...

    def register_component(
        self,
        component_type: type[ComponentT],
        /,
        *,
        identifier: str,
    ) -> type[ComponentT]:
        r"""Register a component to this component manager.

        This returns the provided class, such that this method can serve as a
        decorator.

        Parameters
        ----------
        component_type
            The component class to register.
        identifier
            The identifier with which to register this component class.

        Returns
        -------
        :class:`type`\[:data:`.ComponentT`]
            The component class that was just registered.

        """
        ...

    def deregister_component(self, identifier: str, /) -> None:
        """Deregister a component from this component manager.

        After deregistration, the component will no be tracked, and its
        callbacks can no longer fire until it is re-registered.

        Parameters
        ----------
        identifier
            The identifier of the component class to deregister.

        Returns
        -------
        type[RichComponent]
            The component class that was just deregistered.

        """

    def add_to_client(self, client: disnake.Client, /) -> None:
        """Register this manager to the provided client.

        This is required to make components registered to this manager
        responsive.

        This method registers the :meth:`invoke` callback as an event to the
        client for the :obj:`disnake.on_message_interaction` and
        :obj:`disnake.on_modal_submit` events.

        .. note::
            There is no need to separately register every manager you make.
            In general, it is sufficient to only register the root manager as
            the root manager will contain all components of its children.
            That is, the root manager contains *all* registered components, as
            every other manager is a child of the root manager.

        Parameters
        ----------
        client
            The client to which to register this manager.

        Raises
        ------
        RuntimeError
            This manager has already been registered to the provided client.

        """
        ...

    def remove_from_client(self, client: disnake.Client, /) -> None:
        """Deregister this manager from the provided client.

        This makes all components registered to this manager unresponsive.

        Parameters
        ----------
        client
            The client from which to deregister this manager.

        Raises
        ------
        RuntimeError
            This manager is not registered to the provided client.

        """
        ...

    async def invoke_component(
        self,
        interaction: disnake.MessageInteraction[disnake.Client],
        /,
    ) -> None:
        """Try to invoke a component with the given interaction.

        If this manager has no registered component that matches the interaction,
        it is silently ignored. Otherwise, the interaction will be parsed into
        a fully fledged component, and its callback will then be invoked.

        Parameters
        ----------
        interaction
            The interaction with which to try to invoke a component callback.

        """
        ...


class ComponentFactory(typing.Protocol[ComponentT]):
    """The baseline protocol for any kind of component factory.

    Any and all component factories must implement this protocol in order to be
    properly handled by disnake-compass.

    A component factory handles creating a component instance from a custom id
    by running all individual fields' parsers and aggregating the result into
    a component instance.
    """

    __slots__: typing.Sequence[str] = ()

    @classmethod
    def from_component(
        cls,
        component: type[ComponentT],
        /,
    ) -> typing_extensions.Self:
        """Create a component factory from the provided component.

        This takes the component's fields into account and generates the
        corresponding parser types for each field if a parser was not provided
        manually for that particular field.

        Parameters
        ----------
        component
            The component for which to create a component factory.

        """
        ...

    # TODO: Update docstring
    async def load_params(
        self,
        params: typing.Sequence[str],
        /,
    ) -> typing.Mapping[str, object]:
        """Create a new component instance from the provided custom id.

        This requires the custom id to already have been decomposed into
        individual fields. This is generally done using the
        :meth:`ComponentManager.get_identifier` method.

        Parameters
        ----------
        params
            A mapping of field name to to-be-parsed field values.

        """
        # TODO: Return an ext-components specific conversion error.
        ...

    async def dump_params(self, component: ComponentT, /) -> typing.Mapping[str, str]:
        """Dump a component into a new custom id string.

        This converts the component's individual fields back into strings and
        and uses these strings to build a new custom id. This is generally done
        using the :meth:`ComponentManager.get_identifier` method.

        Parameters
        ----------
        component
            The component to dump into a custom id.

        """
        ...

    async def build_component(
        self,
        params: typing.Sequence[str],
        component_params: typing.Mapping[str, object] | None,
    ) -> ComponentT:
        """Create a new component instance from the provided interaction.

        This requires the custom id to already have been decomposed into
        individual fields. This is generally done by the component manager.

        Parameters
        ----------
        params
            A sequence of to-be-parsed field values.
        component_params
            A mapping of parameters that is to be directly passed to the
            component constructor.

        """
        ...



"""An example showcasing how attrs utilities can be used with disnake-compass."""

import os

import disnake
import disnake_compass
from disnake.ext import commands

bot = commands.InteractionBot()

manager = disnake_compass.get_manager()
manager.add_to_client(bot)


@manager.register
class CustomisableSelect(disnake_compass.RichStringSelect):
    def __attrs_post_init__(self) -> None:
        self.max_values = len(self.options)

    async def callback(
        self, interaction: disnake.MessageInteraction[disnake.Client]
    ) -> None:
        selection = (
            "\n".join(f"- {value}" for value in interaction.values)
            if interaction.values
            else "nothing :("
        )

        await interaction.response.send_message(
            f"You selected:\n{selection}",
            ephemeral=True,
        )


@bot.slash_command()
async def make_select(
    interaction: disnake.CommandInteraction[disnake.Client], options: str
) -> None:
    if not options.strip():
        await interaction.response.send_message("You must specify at least one option!")
        return

    actual_options = [
        disnake.SelectOption(label=option.strip())
        for option in options.split(",")
    ]  # fmt: skip

    if len(actual_options) > 25:
        await interaction.response.send_message("You must specify at most 25 options!")
        return

    component = await CustomisableSelect(options=actual_options).as_ui_component()
    await interaction.response.send_message(
        components=component,
    )


bot.run(os.getenv("EXAMPLE_TOKEN"))



# pyright: reportImportCycles = false
# pyright: reportWildcardImportFromLibrary = false
# ^ This is a false positive as it is confused with site-packages' disnake.

"""The main disnake-compass module.

An extension for disnake aimed at making component interactions with
listeners somewhat less cumbersome.
"""

__title__ = "disnake-compass"
__author__ = "Sharp-Eyes"
__license__ = "MIT"
__copyright__ = "Copyright 2023-present Sharp-Eyes"
__version__ = "1.1.0"

from disnake import VersionInfo as _VersionInfo

from disnake_compass import api as api
from disnake_compass import internal as internal
from disnake_compass.fields import *
from disnake_compass.impl import *

version_info: _VersionInfo = _VersionInfo(major=1, minor=1, micro=0, releaselevel="final", serial=0)

del _VersionInfo



"""Parser implementations for types provided in the datetime package."""

import datetime
import enum
import typing

import attrs

from disnake_compass.impl.parser import base as parser_base
from disnake_compass.impl.parser import builtins as builtins_parsers

__all__: typing.Sequence[str] = (
    "DateParser",
    "DatetimeParser",
    "TimeParser",
    "TimedeltaParser",
    "TimezoneParser",
)

_VALID_BASE_10 = frozenset([10**i for i in range(-6, 0)])


class Resolution(float, enum.Enum):
    r"""The resolution with which :class:`datetime.datetime`\s etc. are stored."""

    MICROS = 1e-6
    """Microsecond resolution.

    This is the default for the datetime module, but often more than required.
    """
    MILLIS = 1e-3
    """Millisecond resolution.

    Rounds the datetime **down** to the nearest microsecond.
    """
    SECONDS = 1
    """Second resolution.

    Rounds the datetime **down** to the nearest second.
    """
    MINUTES = 60 * SECONDS
    """Minute resolution.

    Rounds the datetime **down** to the nearest minute.
    """
    HOURS = 60 * MINUTES
    """Hour resolution.

    Rounds the datetime **down** to the nearest hour.
    """
    DAYS = 24 * HOURS
    """Day resolution.

    Rounds the datetime **down** to the nearest day.
    """


def _validate_resolution(
    _instance: object,
    _attribute: object,
    resolution: float,
) -> None:
    if resolution < Resolution.MICROS:
        msg = f"Resolution must be greater than 1e-6, got {resolution}."
        raise ValueError(msg)

    if resolution < 1 and resolution not in _VALID_BASE_10:
        # TODO: Verify whether this doesn't false-negative
        msg = f"Resolutions smaller than 1 must be a power of 10, got {resolution}."
        raise ValueError(msg)


# TODO: Is forcing the use of timezones on users really a parser_based move?
#       Probably.
@parser_base.register_parser_for(datetime.datetime)
@attrs.define(slots=True)
class DatetimeParser(parser_base.Parser[datetime.datetime]):
    r"""Parser type with support for datetimes.

    Parameters
    ----------
    int_parser:
        The :class:`IntParser` to use internally for this parser.
    resolution:
        The resolution with which to store :class:`~datetime.datetime`\s in custom ids.
        Defaults to :obj:`Resolution.SECONDS`.
    timezone:
        The timezone to use for parsing.
        Defaults to :obj:`datetime.timezone.utc`.
    strict:
        Whether this parser is in strict mode.
        Defaults to ``True``.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`IntParser` to use internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default datetime parser will also return compressed results.
    """
    resolution: int | float = attrs.field(
        default=Resolution.SECONDS,
        validator=_validate_resolution,
        kw_only=True,
    )
    r"""The resolution with which to store :class:`~datetime.datetime`\s in seconds.

    .. warning::
        The resolution must be greater than ``1e-6``, and if the resolution is
        smaller than 1, it **must** be a power of 10. If the resolution is
        greater than 1, it is coerced into an integer.

    .. note::
        Python datetime objects have microsecond accuracy. For most
        applications, this is much more precise than necessary.
        Since custom id space is limited, seconds was chosen as the default.
    """
    timezone: datetime.timezone = attrs.field(default=datetime.timezone.utc, kw_only=True)
    """The timezone to use for parsing.
    Datetimes returned by :meth:`loads` will always be of this timezone.

    This is *not* stored in the custom id.
    """
    strict: bool = attrs.field(default=True, kw_only=True)
    """Whether the parser is in strict mode.

    If the parser is in strict mode, :meth:`loads` requires the provided
    datetime object to be of the correct :attr:`timezone`.
    """

    async def loads(self, argument: str, /) -> datetime.datetime:
        """Load a datetime from a string.

        This uses the underlying :attr:`int_parser`.

        The returned datetime is always of the specified :attr:`timezone`.

        Parameters
        ----------
        argument:
            The string that is to be converted into a datetime.

        """
        return datetime.datetime.fromtimestamp(
            await self.int_parser.loads(argument) * self.resolution,
            tz=self.timezone,
        )

    async def dumps(self, argument: datetime.datetime, /) -> str:
        """Dump a datetime into a string.

        This uses the underlying :attr:`int_parser`.

        If :attr:`strict` is set to ``True``, this will fail if the provided
        ``argument`` does not have a timezone set. Otherwise, a timezone-naive
        datetime will automatically get its timezone set to :attr:`timezone`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        Raises
        ------
        :class:`ValueError`:
            Either the parser is set to strict and the provided datetime was
            timezone-naive, or the provided datetime's timezone does not match
            that of the parser.

        """
        if self.strict:
            if not argument.tzinfo:
                msg = "Strict DatetimeParsers can only load timezone-aware datetimes."
                raise ValueError(msg)
        else:
            argument = argument.replace(tzinfo=self.timezone)

        if argument.tzinfo != self.timezone:
            msg = (
                "Cannot dump the provided datetime object due to a mismatch in"
                f" timezones. (expected: {self.timezone}, got: {argument.tzinfo})"
            )
            raise ValueError(msg)

        timestamp = argument.timestamp()
        if self.resolution != 0:
            timestamp //= self.resolution

        return await self.int_parser.dumps(int(timestamp))


@parser_base.register_parser_for(datetime.timedelta)
@attrs.define(slots=True)
class TimedeltaParser(parser_base.Parser[datetime.timedelta]):
    r"""Parser type with support for :class:`datetime.timedelta`\s.

    Parameters
    ----------
    int_parser:
        The :class:`IntParser` to use internally for this parser.
    resolution:
        The resolution with which to store :class:`~datetime.timedelta`\s in custom ids.
        Defaults to :obj:`Resolution.SECONDS`.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`IntParser` to use internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default datetime parser will also return compressed results.
    """
    resolution: int | float = attrs.field(
        default=Resolution.SECONDS,
        validator=_validate_resolution,
        kw_only=True,
    )
    r"""The resolution with which to store :class:`~datetime.timedelta`\s in seconds.

    .. warning::
        The resolution must be greater than ``1e-6``, and if the resolution is
        smaller than 1, it **must** be a power of 10. If the resolution is
        greater than 1, it is coerced into an integer.

    .. note::
        Python datetime objects have microsecond accuracy. For most
        applications, this is much more precise than necessary.
        Since custom id space is limited, seconds was chosen as the default.
    """

    async def loads(self, argument: str, /) -> datetime.timedelta:
        """Load a timedelta from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The string that is to be converted into a timedelta.

        """
        seconds = await self.int_parser.loads(argument) * self.resolution
        return datetime.timedelta(seconds=seconds)

    async def dumps(self, argument: datetime.timedelta, /) -> str:
        """Dump a timedelta into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(int(argument.total_seconds() // self.resolution))


@parser_base.register_parser_for(datetime.date)
@attrs.define(slots=True)
class DateParser(parser_base.Parser[datetime.date]):
    """Parser type with support for dates.

    Parameters
    ----------
    int_parser:
        The :class:`IntParser` to use internally for this parser.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`IntParser` to use internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default date parser will also return compressed results.
    """

    async def loads(self, argument: str, /) -> datetime.date:
        """Load a date from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The string that is to be converted into a date.

        """
        return datetime.date.fromordinal(await self.int_parser.loads(argument))

    async def dumps(self, argument: datetime.date, /) -> str:
        """Dump a datetime into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(datetime.date.toordinal(argument))


@parser_base.register_parser_for(datetime.time)
@attrs.define(slots=True)
class TimeParser(parser_base.Parser[datetime.time]):
    r"""Parser type with support for times.

    .. important::
        Unlike :class:`DatetimeParser` etc., resolution for this class is set
        via the underlying :attr:`timedelta_parser`. Note that this class *does*
        proxy it through the :attr:`precision` property, which supports both
        getting and setting.

    Parameters
    ----------
    timedelta_parser:
        The :class:`TimedeltaParser` to use internally for this parser.
    strict:
        Whether this parser is in strict mode.
        Defaults to ``True``.
    timezone:
        The timezone to use for parsing.
        Defaults to :obj:`datetime.timezone.utc`.

    """

    timedelta_parser: TimedeltaParser = attrs.field(
        factory=lambda: TimedeltaParser.default(datetime.timedelta),
    )
    """The :class:`TimedeltaParser` to use internally for this parser.

    Since the default timedelta parser uses base-36 to "compress" numbers, the
    default datetime parser will also return compressed results.
    """
    strict: bool = attrs.field(default=True, kw_only=True)
    """Whether the parser is in strict mode.

    If the parser is in strict mode, :meth:`loads` requires the provided
    datetime object to be of the correct :attr:`timezone`.
    """
    timezone: datetime.timezone = attrs.field(default=datetime.timezone.utc, kw_only=True)
    """The timezone to use for parsing.
    Times returned by :meth:`loads` will always be of this timezone.

    This is *not* stored in the custom id.
    """

    @property
    def resolution(self) -> int | float:
        r"""The resolution with which to store :class:`~datetime.time`\s in seconds.

        .. warning::
            The resolution must be greater than ``1e-6``, and if the resolution is
            smaller than 1, it **must** be a power of 10. If the resolution is
            greater than 1, it is coerced into an integer.

        .. note::
            Python time objects have microsecond accuracy. For most
            applications, this is much more precise than necessary.
            Since custom id space is limited, seconds was chosen as the default.
        """
        return self.timedelta_parser.resolution

    @resolution.setter
    def resolution(self, resolution: float) -> None:
        self.timedelta_parser.resolution = resolution

    async def loads(self, argument: str, /) -> datetime.time:
        """Load a time from a string.

        This uses the underlying :attr:`timedelta_parser`.

        The returned time is always of the specified :attr:`timezone`.

        Parameters
        ----------
        argument:
            The string that is to be converted into a time.

        """
        dt = datetime.datetime.min + await self.timedelta_parser.loads(argument)
        return dt.time().replace(tzinfo=self.timezone)

    async def dumps(self, argument: datetime.time, /) -> str:
        """Dump a time into a string.

        This uses the underlying :attr:`timedelta_parser`.

        If :attr:`strict` is set to ``True``, this will fail if the provided
        ``argument`` does not have a timezone set. Otherwise, a timezone-naive
        time will automatically get its timezone set to :attr:`timezone`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        Raises
        ------
        :class:`ValueError`:
            Either the parser is set to strict and the provided time was
            timezone-naive, or the provided time's timezone does not match
            that of the parser.

        """
        if self.strict:
            if not argument.tzinfo:
                msg = "Strict TimeParsers can only load timezone-aware times."
                raise ValueError(msg)
        else:
            argument = argument.replace(tzinfo=self.timezone)

        if argument.tzinfo != self.timezone:
            msg = (
                "Cannot dump the provided time object due to a mismatch in"
                f" timezones. (expected: {self.timezone}, got: {argument.tzinfo})"
            )
            raise ValueError(msg)

        return await self.timedelta_parser.dumps(
            datetime.timedelta(
                hours=argument.hour,
                minutes=argument.minute,
                seconds=argument.second,
                microseconds=argument.microsecond,
            ),
        )


@parser_base.register_parser_for(datetime.timezone)
@attrs.define(slots=True)
class TimezoneParser(parser_base.Parser[datetime.timezone]):
    r"""Parser type with support for :class:`~datetime.timezone`\s.

    .. important::
        Unlike :class:`DatetimeParser` etc., resolution for this class is set
        via the underlying :attr:`timedelta_parser`. Note that this class *does*
        proxy it through the :attr:`precision` property, which supports both
        getting and setting.

    Parameters
    ----------
    timedelta_parser:
        The :class:`TimedeltaParser` to use internally for this parser.

    """

    timedelta_parser: TimedeltaParser = attrs.field(
        factory=lambda: TimedeltaParser.default(datetime.timedelta),
    )
    """The :class:`TimedeltaParser` to use internally for this parser.

    Since the default timedelta parser uses base-36 to "compress" numbers, the
    default datetime parser will also return compressed results.
    """

    @property
    def resolution(self) -> int | float:
        r"""The resolution with which to store :class:`~datetime.time`\s in seconds.

        .. warning::
            The resolution must be greater than ``1e-6``, and if the resolution is
            smaller than 1, it **must** be a power of 10. If the resolution is
            greater than 1, it is coerced into an integer.

        .. note::
            Python time objects have microsecond accuracy. For most
            applications, this is much more precise than necessary.
            Since custom id space is limited, seconds was chosen as the default.
        """
        return self.timedelta_parser.resolution

    @resolution.setter
    def resolution(self, resolution: float) -> None:
        self.timedelta_parser.resolution = resolution

    async def loads(self, argument: str, /) -> datetime.timezone:
        """Load a timezone from a string.

        This uses the underlying :attr:`timedelta_parser`.

        Parameters
        ----------
        argument:
            The string that is to be converted into a timezone.

        """
        return datetime.timezone(await self.timedelta_parser.loads(argument))

    async def dumps(self, argument: datetime.timezone, /) -> str:
        """Dump a timezone into a string.

        This uses the underlying :attr:`timedelta_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.timedelta_parser.dumps(argument.utcoffset(None))



"""A simple example on the use of component managers with disnake-compass."""

import os

import disnake
import disnake_compass
from disnake.ext import commands

bot = commands.InteractionBot()

manager = disnake_compass.get_manager()
manager.add_to_client(bot)

foo_manager = disnake_compass.get_manager("foo")
deeply_nested_manager = disnake_compass.get_manager("foo.bar.baz")


@foo_manager.register
class FooButton(disnake_compass.RichButton):
    label: str | None = "0"

    count: int

    async def callback(
        self, interaction: disnake.MessageInteraction[disnake.Client]
    ) -> None:
        self.count += 1
        self.label = str(self.count)

        component = await self.as_ui_component()
        await interaction.response.edit_message(components=component)


@deeply_nested_manager.register
class FooBarBazButton(disnake_compass.RichButton):
    label: str | None = "0"

    count: int

    async def callback(
        self, interaction: disnake.MessageInteraction[disnake.Client]
    ) -> None:
        self.count += 1
        self.label = str(self.count)

        component = await self.as_ui_component()
        await interaction.response.edit_message(components=component)


@manager.as_callback_wrapper
async def wrapper(
    manager: disnake_compass.ComponentManager,
    component: disnake_compass.api.RichComponent,
    interaction: disnake.Interaction[disnake.Client],
):
    print(
        f"User {interaction.user.name!r} interacted with component {type(component).__name__!r}...",
    )

    yield

    print(
        f"User {interaction.user.name!r}s interaction with component"
        f" {type(component).__name__!r} was successful!",
    )


class InvalidUserError(Exception):
    def __init__(self, message: str, user: disnake.User | disnake.Member) -> None:
        super().__init__(message)
        self.message = message
        self.user = user


@deeply_nested_manager.as_callback_wrapper
async def check_wrapper(
    manager: disnake_compass.api.ComponentManager,
    component: disnake_compass.api.RichComponent,
    interaction: disnake.Interaction[disnake.Client],
):
    if (
        isinstance(interaction, disnake.MessageInteraction)
        and interaction.message.interaction
        and interaction.user != interaction.message.interaction.user
    ):
        message = "You are not allowed to use this component."
        raise InvalidUserError(message, interaction.user)

    yield


@deeply_nested_manager.as_exception_handler
async def error_handler(
    manager: disnake_compass.ComponentManager,
    component: disnake_compass.api.RichComponent,
    interaction: disnake.Interaction[disnake.Client],
    exception: Exception,
):
    if isinstance(exception, InvalidUserError):
        message = f"{exception.user.mention}, {exception.message}"
        await interaction.response.send_message(message, ephemeral=True)
        return True

    return False


@bot.slash_command()
async def test_button(interaction: disnake.CommandInteraction[disnake.Client]) -> None:
    component = await FooButton(count=0).as_ui_component()
    await interaction.response.send_message(components=component)


@bot.slash_command()
async def test_nested_button(
    interaction: disnake.CommandInteraction[disnake.Client],
) -> None:
    component = await FooBarBazButton(count=0).as_ui_component()
    await interaction.response.send_message(components=component)


bot.run(os.getenv("EXAMPLE_TOKEN"))



"""CI script to bump version for a release."""

from __future__ import annotations

import argparse
import re
import sys
from enum import Enum
from pathlib import Path
from typing import NamedTuple, NoReturn

TARGET_FILE = Path("src/disnake_compass/__init__.py")
ORIG_INIT_CONTENTS = TARGET_FILE.read_text("utf-8")

version_re = re.compile(r"(\d+)\.(\d+)\.(\d+)(?:(a|b|rc)(\d+)?)?")


class _ReleaseLevel(Enum):
    alpha = "a"
    beta = "b"
    candidate = "rc"
    final = ""


class _VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: _ReleaseLevel
    serial: int

    @classmethod
    def from_str(cls, s: str) -> _VersionInfo:
        match = version_re.fullmatch(s)
        if not match:
            msg = f"invalid version: '{s}'"
            raise ValueError(msg)

        major, minor, micro, releaselevel, serial = match.groups()
        return _VersionInfo(
            int(major),
            int(minor),
            int(micro),
            _ReleaseLevel(releaselevel or ""),
            int(serial or 0),
        )

    def __str__(self) -> str:
        s = f"{self.major}.{self.minor}.{self.micro}"
        if self.releaselevel is not _ReleaseLevel.final:
            s += self.releaselevel.value
            if self.serial:
                s += str(self.serial)
        return s

    def to_versioninfo(self) -> str:
        return (
            f"_VersionInfo(major={self.major}, minor={self.minor}, micro={self.micro}, "
            f'releaselevel="{self.releaselevel.name}", serial={self.serial})'
        )


def _get_current_version() -> _VersionInfo:
    match = re.search(r"^__version__\b.*\"(.+?)\"$", ORIG_INIT_CONTENTS, re.MULTILINE)
    assert match, "could not find current version in __init__.py"
    return _VersionInfo.from_str(match[1])


def _replace_line(text: str, regex: str, repl: str) -> str:
    lines: list[str] = []
    found = False

    for line in text.split("\n"):
        if re.search(regex, line):
            found = True
            line = repl  # noqa: PLW2901
        lines.append(line)

    assert found, f"failed to find `{regex}` in file"
    return "\n".join(lines)


def _fail(msg: str) -> NoReturn:
    print("error:", msg, file=sys.stderr)
    sys.exit(1)


def _main() -> None:
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--set", metavar="VERSION", help="set new version (e.g. '1.2.3' or 'dev')")
    group.add_argument("--show", action="store_true", help="print current version")
    args = parser.parse_args()

    current_version = _get_current_version()

    if args.show:
        print(str(current_version))
        return

    # else, update to specified version
    new_version_str = args.set

    if new_version_str == "dev":
        if current_version.releaselevel is not _ReleaseLevel.final:
            _fail("Current version must be final to bump to dev version")
        new_version = _VersionInfo(
            major=current_version.major,
            minor=current_version.minor + 1,
            micro=0,
            releaselevel=_ReleaseLevel.alpha,
            serial=0,
        )
    else:
        new_version = _VersionInfo.from_str(new_version_str)

    text = ORIG_INIT_CONTENTS
    text = _replace_line(text, r"^__version__\b", f'__version__ = "{new_version!s}"')
    text = _replace_line(
        text, r"^version_info\b", f"version_info: _VersionInfo = {new_version.to_versioninfo()}"
    )

    if text != ORIG_INIT_CONTENTS:
        TARGET_FILE.write_text(text, "utf-8")

    print(str(new_version))


if __name__ == "__main__":
    _main()



"""An example of having multiple disnake-compass components interact."""

import os
import typing

import disnake
from disnake.ext import commands

import disnake_compass

DEFAULT_OPTION = disnake.SelectOption(
    label="Please enable some options.",
    value="placeholder",
    default=True,
)

bot = commands.InteractionBot()

manager = disnake_compass.get_manager()
manager.add_to_client(bot)


@manager.register()
class OptionsToggleButton(disnake_compass.RichButton):
    """A button component that enables/disables options on a DynamicSelectMenu."""

    style: disnake.ButtonStyle = disnake.ButtonStyle.red
    options: typing.Sequence[str]

    def parse_options(self) -> typing.Sequence[disnake.SelectOption]:
        if self.style == disnake.ButtonStyle.red:
            return []

        return [disnake.SelectOption(label=option) for option in self.options]

    def update_select(self, components: typing.Sequence[disnake_compass.api.RichComponent]):
        select: DynamicSelectMenu | None = None
        options: list[disnake.SelectOption] = []

        for component in components:
            if isinstance(component, DynamicSelectMenu):
                if select is not None:
                    raise RuntimeError("Found more than one DynamicSelectMenu.")

                select = component

            elif isinstance(component, OptionsToggleButton):
                options.extend(component.parse_options())

        if not select:
            raise RuntimeError("Could not find a DynamicSelectMenu.")

        select.set_options(options)

    async def callback(self, interaction: disnake.MessageInteraction[disnake.Client]):
        # Get all components on the message for easier re-sending.
        # Both of these lists will automagically contain self so that any
        # changes immediately reflect without extra effort.
        layout, components = await manager.parse_message_components(interaction.message.components)

        # Toggle style for the clicked button.
        self.style = (
            disnake.ButtonStyle.red
            if self.style == disnake.ButtonStyle.green
            else disnake.ButtonStyle.green
        )

        # Add/remove options to the DynamicSelect based on whether this
        # button was toggled on or off.
        self.update_select(components)

        # Re-send and update all components.
        await manager.update_layout(layout, components)
        await interaction.response.edit_message(components=layout)


@manager.register()
class DynamicSelectMenu(disnake_compass.RichStringSelect):
    """A select menu that has its options externally managed."""

    def __attrs_post_init__(self) -> None:  # See the `attrs.py` example.
        self.set_options([])

    def set_options(self, options: typing.List[disnake.SelectOption]):
        if options:
            self.options = options
            self.max_values = len(options)
            self.disabled = False

        else:
            self.options = [DEFAULT_OPTION]
            self.max_values = 1
            self.disabled = True

    async def callback(self, interaction: disnake.MessageInteraction[disnake.Client]) -> None:
        selection = (
            "\n".join(f"- {value}" for value in interaction.values)
            if interaction.values
            else "nothing :("
        )

        await interaction.response.send_message(f"You selected:\n{selection}")


@bot.slash_command()
async def test_components(
    interaction: disnake.CommandInteraction[disnake.Client],
) -> None:
    # TODO: Maybe a utility method to turn a sequence of rich components into a
    #       sequence of UI components?
    layout = [
        [
            await OptionsToggleButton(
                label="numbers", options=["1", "2", "3", "4", "5"]
            ).as_ui_component(),
            await OptionsToggleButton(
                label="letters", options=["a", "b", "c", "d", "e"]
            ).as_ui_component(),
            await OptionsToggleButton(
                label="symbols", options=["*", "&", "#", "+", "-"]
            ).as_ui_component(),
        ],
        [await DynamicSelectMenu().as_ui_component()],
    ]

    await interaction.response.send_message(components=layout)


bot.run(os.environ["EXAMPLE_TOKEN"])



"""Re-exports of some disnake internal types related to components."""

import typing

import disnake

# NOTE: Keep up to date with disnake types...

ActionRowMessageComponent = disnake.ui.Button[typing.Any] | disnake.ui.Select[typing.Any]
MessageTopLevelComponentV2 = (
    disnake.ui.Section
    | disnake.ui.TextDisplay
    | disnake.ui.MediaGallery
    | disnake.ui.File
    | disnake.ui.Separator
    | disnake.ui.Container
)
ModalTopLevelComponent_ = disnake.ui.TextDisplay | disnake.ui.Label
ActionRowChildT = typing.TypeVar("ActionRowChildT", bound=disnake.ui.WrappedComponent)
NonActionRowChildT = typing.TypeVar(
    "NonActionRowChildT",
    bound=MessageTopLevelComponentV2 | ModalTopLevelComponent_,
)
AnyUIComponentInput = ActionRowChildT | disnake.ui.ActionRow[ActionRowChildT] | NonActionRowChildT
ComponentInput = (
    AnyUIComponentInput[ActionRowChildT, NonActionRowChildT]
    | typing.Sequence[
        AnyUIComponentInput[ActionRowChildT, NonActionRowChildT] | typing.Sequence[ActionRowChildT]
    ]
)
MessageComponents = ComponentInput[ActionRowMessageComponent, MessageTopLevelComponentV2]



"""A simple example on the use of selects with disnake-compass."""

from __future__ import annotations

import os
import typing

import disnake
import disnake_compass
from disnake.ext import commands

bot = commands.InteractionBot()

manager = disnake_compass.get_manager()
manager.add_to_client(bot)


LEFT = "\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}"
MIDDLE = "\N{BLACK CIRCLE FOR RECORD}\N{VARIATION SELECTOR-16}"
RIGHT = "\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}"

SLOT_OPTIONS = [
    disnake.SelectOption(label="Left", value="left", emoji=LEFT),
    disnake.SelectOption(label="Middle", value="middle", emoji=MIDDLE),
    disnake.SelectOption(label="Right", value="right", emoji=RIGHT),
    disnake.SelectOption(label="Finalise", emoji="\N{WHITE HEAVY CHECK MARK}"),
]


BLACK_SQUARE = "\N{BLACK LARGE SQUARE}"
BLUE_SQUARE = "\N{LARGE BLUE SQUARE}"
BROWN_SQUARE = "\N{LARGE BROWN SQUARE}"
GREEN_SQUARE = "\N{LARGE GREEN SQUARE}"
PURPLE_SQUARE = "\N{LARGE PURPLE SQUARE}"
RED_SQUARE = "\N{LARGE RED SQUARE}"
WHITE_SQUARE = "\N{WHITE LARGE SQUARE}"
YELLOW_SQUARE = "\N{LARGE YELLOW SQUARE}"

COLOUR_OPTIONS = [
    disnake.SelectOption(label="Black", value=BLACK_SQUARE, emoji=BLACK_SQUARE),
    disnake.SelectOption(label="Blue", value=BLUE_SQUARE, emoji=BLUE_SQUARE),
    disnake.SelectOption(label="Brown", value=BROWN_SQUARE, emoji=BROWN_SQUARE),
    disnake.SelectOption(label="Green", value=GREEN_SQUARE, emoji=GREEN_SQUARE),
    disnake.SelectOption(label="Purple", value=PURPLE_SQUARE, emoji=PURPLE_SQUARE),
    disnake.SelectOption(label="Red", value=RED_SQUARE, emoji=RED_SQUARE),
    disnake.SelectOption(label="White", value=WHITE_SQUARE, emoji=WHITE_SQUARE),
    disnake.SelectOption(label="Yellow", value=YELLOW_SQUARE, emoji=YELLOW_SQUARE),
]


@manager.register
class MySelect(disnake_compass.RichStringSelect):
    placeholder: str | None = "Please select a square."
    options: typing.List[disnake.SelectOption] = SLOT_OPTIONS

    slot: str = "0"
    state: str = "slot"
    colour_left: str = BLACK_SQUARE
    colour_middle: str = BLACK_SQUARE
    colour_right: str = BLACK_SQUARE

    async def callback(
        self, interaction: disnake.MessageInteraction[disnake.Client]
    ) -> None:
        assert interaction.values is not None
        selected = interaction.values[0]

        if self.state == "slot":
            self.handle_slots(selected)

        else:
            self.handle_colours(selected)

        msg = self.render_colours()
        component = await self.as_ui_component()
        await interaction.response.edit_message(msg, components=component)

    def handle_slots(self, selected: str) -> None:
        if selected == "Finalise":
            self.disabled = True
            self.placeholder = "Woo!"
            return

        self.options = COLOUR_OPTIONS
        self.placeholder = f"Please select a colour for the {selected} square."

        self.slot = selected
        self.state = "colour"

    def handle_colours(self, selected: str) -> None:
        self.options = SLOT_OPTIONS

        setattr(self, f"colour_{self.slot}", selected)
        self.state = "slot"

    def render_colours(self) -> str:
        return f"{self.colour_left}{self.colour_middle}{self.colour_right}\n"


@bot.slash_command()
async def test_select(interaction: disnake.CommandInteraction[disnake.Client]) -> None:
    my_select = MySelect()
    await interaction.response.send_message(
        my_select.render_colours(),
        components=await my_select.as_ui_component(),
    )


bot.run(os.getenv("EXAMPLE_TOKEN"))



"""Sphinx attributetable extension."""

# pyright: reportUnknownMemberType = false

from __future__ import annotations

import asyncio
import collections
import importlib
import inspect
import re
import typing

import typing_extensions
from docutils import nodes
from sphinx import addnodes
from sphinx.locale import _
from sphinx.util.docutils import SphinxDirective

if typing.TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment
    from sphinx.writers.html import HTMLTranslator


class SphinxExtensionMeta(typing.TypedDict, total=False):
    version: str
    env_version: int
    parallel_read_safe: bool
    parallel_write_safe: bool


class attributetable(nodes.General, nodes.Element):
    pass


class attributetablecolumn(nodes.General, nodes.Element):
    pass


class attributetabletitle(nodes.Referential, nodes.TextElement):
    pass


class attributetableplaceholder(nodes.General, nodes.Element):
    pass


class attributetablebadge(nodes.TextElement):
    pass


class attributetable_item(nodes.Part, nodes.Element):
    pass


def visit_attributetable_node(self: HTMLTranslator, node: nodes.Element) -> None:
    class_ = typing.cast(str, node["python-class"])
    self.body.append(f'<div class="py-attribute-table" data-move-to-id="{class_}">')


def visit_attributetablecolumn_node(self: HTMLTranslator, node: nodes.Element) -> None:
    self.body.append(self.starttag(node, "div", CLASS="py-attribute-table-column"))


def visit_attributetabletitle_node(self: HTMLTranslator, node: nodes.Element) -> None:
    self.body.append(self.starttag(node, "span"))


def visit_attributetablebadge_node(self: HTMLTranslator, node: nodes.Element) -> None:
    """Add a class to each badge of the type that it is."""
    badge_type = typing.cast(str, node["badge-type"])
    if badge_type not in (
        "coroutine",
        "decorator",
        "method",
        "classmethod",
        "attribute",
    ):
        msg = f"badge_type {badge_type} is currently unsupported"
        raise RuntimeError(msg)
    attributes: dict[typing.Literal["class"], str] = {
        "class": f"badge-{badge_type}",
    }
    self.body.append(self.starttag(node, "span", **attributes))


def visit_attributetable_item_node(self: HTMLTranslator, node: nodes.Element) -> None:
    self.body.append(self.starttag(node, "li"))


def depart_attributetable_node(self: HTMLTranslator, _node: nodes.Element) -> None:
    self.body.append("</div>")


def depart_attributetablecolumn_node(
    self: HTMLTranslator,
    _node: nodes.Element,
) -> None:
    self.body.append("</div>")


def depart_attributetabletitle_node(self: HTMLTranslator, _node: nodes.Element) -> None:
    self.body.append("</span>")


def depart_attributetablebadge_node(self: HTMLTranslator, _node: nodes.Element) -> None:
    self.body.append("</span>")


def depart_attributetable_item_node(self: HTMLTranslator, _node: nodes.Element) -> None:
    self.body.append("</li>")


_name_parser_regex = re.compile(r"(?P<module>[\w.]+\.)?(?P<name>\w+)")


class PyAttributeTable(SphinxDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {}  # noqa: RUF012

    def parse_name(self, content: str) -> tuple[str, str | None]:
        match = _name_parser_regex.match(content)
        path, name = match.groups() if match else (None, None)
        if path:
            modulename = path.rstrip(".")
        else:
            modulename = self.env.temp_data.get("autodoc:module")
            if not modulename:
                modulename = self.env.ref_context.get("py:module")
        if modulename is None:
            msg = f"modulename somehow None for {content} in {self.env.docname}."
            raise RuntimeError(msg)

        return modulename, name

    def run(self) -> list[nodes.Node]:
        content = typing.cast(str, self.arguments[0]).strip()
        node = attributetableplaceholder("")
        modulename, name = self.parse_name(content)
        node["python-doc"] = self.env.docname
        node["python-module"] = modulename
        node["python-class"] = name
        node["python-full-name"] = f"{modulename}.{name}"
        return [node]


def build_lookup_table(env: BuildEnvironment) -> dict[str, list[str]]:
    # Given an environment, load up a lookup table of
    # full-class-name: objects
    result: collections.defaultdict[str, list[str]] = collections.defaultdict(list)
    domain = env.domains["py"]

    ignored = {
        "data",
        "exception",
        "module",
        "class",
    }

    for fullname, _unused, objtype, _unused, _unused, _unused in domain.get_objects():
        if objtype in ignored:
            continue

        classname, _unused, child = fullname.rpartition(".")
        result[classname].append(child)

    return result


class TableElement(typing.NamedTuple):
    fullname: str
    label: str
    badge: attributetablebadge | None


def process_attributetable(app: Sphinx, doctree: nodes.document, _docname: str) -> None:
    env = app.builder.env

    lookup = build_lookup_table(env)
    doc_iter = typing.cast(
        typing.Iterator[nodes.document],
        doctree.findall(attributetableplaceholder),
    )
    for node in doc_iter:
        modulename = typing.cast(str, node["python-module"])
        classname = typing.cast(str, node["python-class"])
        fullname = typing.cast(str, node["python-full-name"])
        groups = get_class_results(lookup, modulename, classname, fullname)
        table = attributetable("")
        for label, subitems in groups.items():
            if not subitems:
                continue
            table.append(
                class_results_to_node(
                    _(label),
                    sorted(subitems, key=lambda c: c.label),
                    fullname,
                ),
            )

        table["python-class"] = fullname

        if not table:
            node.replace_self([])
        else:
            node.replace_self([table])


def _is_classvar(ann: str | type) -> bool:
    if isinstance(ann, str):
        return ann.startswith(("typing.ClassVar", "typing_extensions.ClassVar"))

    return typing.get_origin(ann) in (typing.ClassVar, typing_extensions.ClassVar)


def get_class_results(
    lookup: dict[str, list[str]],
    modulename: str,
    name: str,
    fullname: str,
) -> dict[str, list[TableElement]]:
    module = importlib.import_module(modulename)
    cls: type = getattr(module, name)

    groups: dict[str, list[TableElement]] = {
        "Attributes": [],
        "Methods": [],
    }

    try:
        members = lookup[fullname]
    except KeyError:
        return groups

    anns: dict[str, typing.Any] = {}

    for attr in members:
        attrlookup = f"{fullname}.{attr}"
        key = "Attributes"
        badge = None
        label = attr
        value = None

        for base in cls.__mro__[:-1]:
            value = base.__dict__.get(attr)
            anns.update(getattr(base, "__annotations__", {}))
            if value is not None:
                break

        if value is not None:
            doc = value.__doc__ or ""
            if asyncio.iscoroutinefunction(value) or doc.startswith("|coro|"):
                key = "Methods"
                badge = attributetablebadge("async", "async")
                badge["badge-type"] = _("coroutine")
            elif isinstance(value, classmethod):
                key = "Methods"
                badge = attributetablebadge("cls", "cls")
                badge["badge-type"] = "classmethod"
            elif inspect.isfunction(value):
                if doc.lstrip().startswith(("A decorator", "A shortcut decorator")):
                    # finicky but surprisingly consistent
                    badge = attributetablebadge("@", "@")
                    badge["badge-type"] = _("decorator")
                    key = "Methods"
                else:
                    key = "Methods"
                    badge = attributetablebadge("def", "def")
                    badge["badge-type"] = _("method")
            elif attr in anns and _is_classvar(anns[attr]):
                key = "Attributes"
                badge = attributetablebadge("classvar", "classvar")
                badge["badge-type"] = _("attribute")
        elif attr in anns and _is_classvar(anns[attr]):
            key = "Attributes"
            badge = attributetablebadge("classvar", "classvar")
            badge["badge-type"] = _("attribute")

        groups[key].append(TableElement(fullname=attrlookup, label=label, badge=badge))

    return groups


def class_results_to_node(
    key: str,
    elements: list[TableElement],
    fullname: str,
) -> attributetablecolumn:
    titleref = nodes.reference(
        "",
        "",
        nodes.Text(key),
        internal=True,
        refuri=f"#{fullname.replace('.', '-')}-{key}".lower(),
        anchorname="",
    )
    title = attributetabletitle("", "", titleref)
    ul = nodes.bullet_list("")
    ul["classes"].append("py-attribute-table-list")
    for element in elements:
        ref = nodes.reference(
            "",
            "",
            nodes.Text(element.label),
            internal=True,
            refuri="#" + element.fullname,
            anchorname="",
        )
        para = addnodes.compact_paragraph("", "", ref)
        if element.badge is not None:
            ul.append(attributetable_item("", element.badge, para))
        else:
            ul.append(attributetable_item("", para))

    return attributetablecolumn("", title, ul)


def setup(app: Sphinx) -> dict[str, bool]:
    app.add_directive("attributetable", PyAttributeTable)
    app.add_node(
        attributetable,
        html=(visit_attributetable_node, depart_attributetable_node),
    )
    app.add_node(
        attributetablecolumn,
        html=(visit_attributetablecolumn_node, depart_attributetablecolumn_node),
    )
    app.add_node(
        attributetabletitle,
        html=(visit_attributetabletitle_node, depart_attributetabletitle_node),
    )
    app.add_node(
        attributetablebadge,
        html=(visit_attributetablebadge_node, depart_attributetablebadge_node),
    )
    app.add_node(
        attributetable_item,
        html=(visit_attributetable_item_node, depart_attributetable_item_node),
    )
    app.add_node(attributetableplaceholder)
    app.connect("doctree-resolved", process_attributetable)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Implementation of component base classes.

To easily integrate with disnake-compass, it is recommended to inherit
from any of these base classes. In any case, it is very much recommended to at
least use the `ComponentMeta` metaclass. Without this, a lot of internal
functionality will have to be manually re-implemented.
"""

from __future__ import annotations

import abc
import sys
import typing

import attrs
import typing_extensions

from disnake_compass import fields as fields
from disnake_compass.api import component as component_api
from disnake_compass.api import parser as parser_api
from disnake_compass.impl import factory as factory_impl
from disnake_compass.impl import parser as parser_impl

if typing.TYPE_CHECKING:
    import disnake

__all__: typing.Sequence[str] = ("ComponentBase",)


_T = typing.TypeVar("_T")

MaybeCoroutine: typing.TypeAlias = _T | typing.Coroutine[None, None, _T]
_AnyAttr: typing_extensions.TypeAlias = "attrs.Attribute[typing.Any]"


def _is_attrs_pass(namespace: dict[str, typing.Any]) -> bool:
    """Check if attrs has already influenced the class' namespace.

    Note that we check the namespace instead of using `attrs.has`, because
    `attrs.has` would always return `True` for a class inheriting an attrs class,
    and we specifically need to distinguish between the two passes inside
    `ComponentMeta.__new__`.
    """
    return namespace.get("__attrs_attrs__") is not None


def _determine_parser(
    attribute: _AnyAttr,
    overwrite: _AnyAttr | None,
    *,
    required: bool = True,
) -> parser_api.Parser[typing.Any] | None:
    parser = fields.get_parser(attribute)
    if parser:
        return parser

    if overwrite:
        parser = fields.get_parser(overwrite)
        if parser:
            return parser

    if required:
        return parser_impl.get_parser(attribute.type or str)

    return None


def _eval_type(cls: type, annotation: typing.Any) -> typing.Any:  # noqa: ANN401
    # Get the module globals in which the class was defined. This is the most
    # probable candidate in which to find the type annotations' definitions.
    #
    # For the most part, this should be safe. Conflicts where e.g. a component
    # inheriting from RichButton but not defining _AnyEmoji in their own module
    # are safe, because the type has already been passed through this function
    # when the RichButton class was initially created.
    cls_globals = sys.modules[cls.__module__].__dict__

    if isinstance(annotation, str):
        annotation = typing.ForwardRef(annotation, is_argument=False)

    # Evaluate the typehint with the provided globals.
    return typing._eval_type(annotation, cls_globals, None)  # pyright: ignore  # noqa: PGH003, SLF001


def _assert_valid_overwrite(attribute: _AnyAttr, overwrite: _AnyAttr) -> None:
    if fields.FieldMetadata.FIELDTYPE not in overwrite.metadata:
        return

    # The field was defined using fields.field / fields.internal / etc.
    # Ensure the overwrite matches the original field type.

    attribute_type = fields.get_field_type(overwrite)
    overwrite_type = fields.get_field_type(attribute)

    if overwrite_type is not attribute_type:
        new = (attribute_type.name or "<unknown>").lower()
        old = (overwrite_type.name or "<unknown>").lower()
        msg = (
            f"Invalid field override. Field {overwrite.name} is defined as"
            f" a(n) {new} field, but was overwritten as a(n) {old} field."
        )
        raise TypeError(msg)


def _is_custom_id_field(field: _AnyAttr) -> bool:
    return fields.get_field_type(field, fields.FieldType.CUSTOM_ID) is fields.FieldType.CUSTOM_ID


def _field_transformer(
    cls: type,
    attributes: list[_AnyAttr],
) -> list[_AnyAttr]:
    super_attributes: dict[str, _AnyAttr] = (
        {field.name: field for field in fields.get_fields(cls)} if attrs.has(cls) else {}
    )

    finalised_attributes: list[_AnyAttr] = []
    for attribute in attributes:
        super_attribute = super_attributes.get(attribute.name)

        # Ensure all forward-references are evaluated.
        evolved = attribute.evolve(type=_eval_type(cls, attribute.type))

        if super_attribute:
            # This field overwrites a pre-existing field. Merge metadata and
            # ensure the parser isn't overwritten if the new value is None.
            _assert_valid_overwrite(super_attribute, attribute)
            metadata = {**super_attribute.metadata, **evolved.metadata}

        else:
            # Not an overwrite, ensure the fieldtype is set to CUSTOM_ID if not
            # already provided.
            metadata = {
                fields.FieldMetadata.FIELDTYPE: fields.FieldType.CUSTOM_ID,
                **evolved.metadata,
            }

        # TODO: Make copy of parser instead of using the same instance
        metadata[fields.FieldMetadata.PARSER] = _determine_parser(
            evolved,
            super_attribute,
            # Fields only need a parser if
            # - The component is concrete,
            # - The field is a custom-id field.
            # In case of an overwrite, we check the field type of the super-field.
            required=(
                not typing_extensions.is_protocol(cls)
                and _is_custom_id_field(super_attribute or attribute)
            ),
        )

        # Apply finalised metadata.
        finalised_attributes.append(evolved.evolve(metadata=metadata))

    return finalised_attributes


@typing_extensions.dataclass_transform(
    kw_only_default=True,
    field_specifiers=(fields.field, fields.internal, fields.meta),
)
class ComponentMeta(type(typing.Protocol)):
    """Metaclass for all disnake-compass component types.

    It is **highly** recommended to use this metaclass for any class that
    should interface with the components api exposed by
    `disnake-compass`.

    This metaclass handles :mod:`attrs` class generation, custom id completion,
    interfacing with component managers, parser and factory generation, and
    automatic slotting.
    """

    def __new__(  # noqa: PYI034
        metacls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, typing.Any],
    ) -> ComponentMeta:
        # NOTE: This is run twice for each new class; once for the actual class
        #       definition, and once more by attrs.define(). We ensure we only
        #       run the full class creation logic once.

        cls = super().__new__(metacls, name, bases, namespace)

        # If this is attrs' pass, return immediately.
        if _is_attrs_pass(namespace):
            return cls

        cls = attrs.define(cls, slots=True, kw_only=True, field_transformer=_field_transformer)

        # NOTE: Pyright complains about RichComponent being a data protocol
        #       here, but this is a false-positive, as the only non-method
        #       member is __slots__.
        assert issubclass(cls, component_api.RichComponent)  # pyright: ignore[reportGeneralTypeIssues]
        factory_cls = (
            factory_impl.NoopFactory
            if typing_extensions.is_protocol(cls)
            else factory_impl.ComponentFactory[typing.Any]
        )

        cls.set_factory(factory_cls.from_component(cls))
        return typing.cast(ComponentMeta, cls)


@typing.runtime_checkable
class ComponentBase(component_api.RichComponent, typing.Protocol, metaclass=ComponentMeta):
    """Overarching base class for any kind of component."""

    _factory: typing.ClassVar[component_api.ComponentFactory[typing_extensions.Self]]
    _manager: typing.ClassVar[component_api.ComponentManager | None] = None

    @classmethod
    def get_manager(cls) -> component_api.ComponentManager:  # noqa: D102
        # <<Docstring inherited from component_api.RichComponent>>

        if cls._manager is None:
            msg = f"Component {cls.__qualname__} is not yet registered to a manager."
            raise RuntimeError(msg)

        return cls._manager

    @classmethod
    def set_manager(cls, manager: component_api.ComponentManager | None) -> None:  # noqa: D102
        # <<Docstring inherited from component_api.RichComponent>>

        cls._manager = manager

    @classmethod
    def get_factory(cls) -> component_api.ComponentFactory[typing_extensions.Self]:
        r"""Get the factory that built this component instance.

        .. note::
            Component base classes can be declared as a subclass of
            :class:`~typing.Protocol`\s, which tells disnake_compass that this
            component is a 'template' and should not be instantiable.
            In this case, this attribute defaults to a :class:`~NoopFactory`.

            In case a concrete component subclass is created, a matching
            :class:`~ComponentFactory` is automatically generated instead.
        """
        return cls._factory

    @classmethod
    def set_factory(cls, factory: component_api.ComponentFactory[typing_extensions.Self]) -> None:  # noqa: D102
        # <<Docstring inherited from component_api.RichComponent>>

        cls._factory = factory

    async def as_ui_component(  # noqa: D102
        self,
        manager: component_api.ComponentManager,
        /,
    ) -> disnake.ui.WrappedComponent:
        # <<Docstring inherited from component_api.RichComponent>>
        ...

    async def make_custom_id(self, manager: component_api.ComponentManager | None, /) -> str:
        """Make a custom id from this component given its current state.

        The generated custom id will contain the full state of the component,
        such that it be used to entirely reconstruct the component later.

        Because parameter to string conversion supports asynchronous callbacks,
        this has to be an async function instead of e.g. a property.

        .. note::
            As the logic for translating a component to and from a custom id
            resides inside the component manager, the component *must* be
            registered to a manager to use this method.

        Parameters
        ----------
        manager:
            The manager to use to make a custom id for this component. This is
            only relevant if you have multiple managers for the same component.
            Defaults to the root manager.

        Returns
        -------
        str:
            The custom id representing the full state of this component.

        """
        if manager is None:
            manager = self.get_manager()

        return await manager.make_custom_id(self)

    @abc.abstractmethod
    async def callback(  # pyright: ignore[reportIncompatibleMethodOverride]  # noqa: D102
        self,
        inter: disnake.MessageInteraction[disnake.Client],
        /,
    ) -> None:
        # <<docstring inherited from component_api.RichButton>>

        # NOTE: We narrow the interaction type down to a disnake.MessageInteraction
        #       here. This isn't typesafe, but it's just cleaner for the user.
        ...



# pyright: reportImportCycles = false
# pyright: reportWildcardImportFromLibrary = false
# ^ This is a false positive as it is confused with site-packages' disnake.

"""Protocols for disnake-compass internals.

These are provided to decouple the implementation details from their respective
implementations, so that custom implementations can be created accordingly.
This allows users to slot their own implementations into the library that
properly interface with the existing logic, thus maintaining functionality.
"""

from __future__ import annotations

from disnake_compass.api.component import *
from disnake_compass.api.parser import *



"""Sphinx configuration."""

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import re
import subprocess
import sys

import disnake
import sphinx.config
from disnake.ext import commands

import disnake_compass
from docs.source import util


def git(*args: str) -> str:
    """Run a git command and return the output."""
    return subprocess.check_output(["git", *args], text=True).strip()


git_pwd = git("rev-parse", "--show-toplevel")
sys.path.append(os.path.abspath(git_pwd))


sys.modules["commands"] = commands
sys.path.append(os.path.abspath("./extensions"))

project = disnake_compass.__title__
copyright = "2023, Sharp-Eyes"  # noqa: A001
author = disnake_compass.__author__
release = disnake_compass.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    # In sphinx.
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    # External.
    "sphinxcontrib_trio",
    "sphinxcontrib.towncrier.ext",
    "sphinx_inline_tabs",
    "sphinx_copybutton",
    "hoverxref.extension",
    "sphinx_autodoc_typehints",
    # Custom.
    "attributetable",
    "custom_documenter",
]
exclude_patterns = []

pygments_style = "friendly"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_theme_options = {
    "source_repository": "https://github.com/DisnakeCommunity/disnake-compass",
    "source_branch": "master",
    # Taken directly from Furo docs at https://github.com/pradyunsg/furo/blob/main/docs/conf.py.
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/DisnakeCommunity/disnake-compass/",
            "html": """
                    <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                        <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                    </svg>
                """,  # noqa: E501
            "class": "",
        },
    ],
}
html_show_sourcelink = False

html_static_path = ["_static"]
html_css_files = [
    "./css/custom.css",
    "./css/attributetable.css",
    "./css/custom_documenter.css",
]
html_js_files = ["./js/custom.js"]

# -- Intersphinx config -------------------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "disnake": ("https://docs.disnake.dev/en/stable/", None),
    "attrs": ("https://www.attrs.org/en/stable/", None),
}

# -- Autodoc config -----------------------------------------------------------

autodoc_member_order = "groupwise"
autodoc_typehints = "signature"
autodoc_typehints_format = "short"
autodoc_default_options = {"show-inheritance": True, "inherited-members": False}

# -- Hoverxref config ---------------------------------------------------------

hoverx_default_type = "tooltip"
hoverxref_domains = ["py"]
hoverxref_role_types = dict.fromkeys(
    ["ref", "class", "func", "meth", "attr", "exc", "data", "obj"],
    "tooltip",
)
hoverxref_tooltip_theme = ["tooltipster-custom"]
hoverxref_tooltip_lazy = True

# These have to match the keys on intersphinx_mapping, and those projects must
# be hosted on readthedocs.
hoverxref_intersphinx = list(intersphinx_mapping)


# -- Linkcode config ----------------------------------------------------------

repo_url = "https://github.com/DisnakeCommunity/disnake-compass"


def get_git_ref() -> str:
    """Return the current git reference."""
    # Current git reference. Uses branch/tag name if found, otherwise uses commit hash
    git_ref = git("name-rev", "--name-only", "--no-undefined", "HEAD")
    return re.sub(r"^(remotes/[^/]+|tags)/", "", git_ref)


git_ref = get_git_ref()
module_path = util.get_module_path()
linkcode_resolve = util.make_linkcode_resolver(module_path, repo_url, git_ref)


# -- Extlinks config ----------------------------------------------------------

extlinks = {
    "issue": (f"{repo_url}/issues/%s", "#%s"),
    "github": (f"{repo_url}/%s", "%s"),
    "github-blob": (f"{repo_url}/blob/{git_ref}/%s", "%s"),
    "example": (f"{repo_url}/blob/{git_ref}/examples/%s.py", "View on GitHub: %s.py"),
    "attrs": ("https://www.attrs.org/en/stable/%s", "%s"),
}

# -- towncrier config ---------------------------------------------------------

towncrier_draft_autoversion_mode = "draft"
towncrier_draft_include_empty = False  # hides the unreleased indicator if there are no changes
towncrier_draft_working_directory = git_pwd

# -- sphinx-autodoc-typehints config ------------------------------------------

# Apply monkeypatch.
util.apply_autodoc_typehints_patch()

typehints_document_rtype = False
typehints_use_rtype = False
simplify_optional_unions = True
always_use_bars_union = True
typehints_fully_qualified = False

# Customise display for specific types.
aliases: dict[object, str] = {
    # Idk why this is needed, but it is...
    disnake.ButtonStyle: ":class:`~disnake.ButtonStyle`",
}


def typehints_formatter(ann: object, _: sphinx.config.Config) -> str | None:
    """Format typehints."""
    if typehint := aliases.get(ann):
        return typehint

    return None



# pyright: reportImportCycles = false
# pyright: reportWildcardImportFromLibrary = false
# ^ This is a false positive as it is confused with site-packages' disnake.

"""Implementations for all kinds of parser classes."""

from disnake_compass.impl.parser.base import *
from disnake_compass.impl.parser.builtins import *
from disnake_compass.impl.parser.channel import *
from disnake_compass.impl.parser.datetime import *
from disnake_compass.impl.parser.emoji import *
from disnake_compass.impl.parser.enum import *
from disnake_compass.impl.parser.guild import *
from disnake_compass.impl.parser.message import *
from disnake_compass.impl.parser.snowflake import *
from disnake_compass.impl.parser.user import *



"""An example showcasing a simple cv2-style paginator powered by disnake-compass."""

import os

import disnake
import disnake_compass
from disnake.ext import commands

bot = commands.InteractionBot()

manager = disnake_compass.get_manager()
manager.add_to_client(bot)


@manager.register()
class PageAdvanceButton(disnake_compass.RichButton):
    increment: int

    async def callback(self, inter: disnake.MessageInteraction[disnake.Client]) -> None:
        _, components = await manager.parse_message_components(inter.message.components)

        # Find and increment the page tracker...
        for component in components:
            if isinstance(component, PageTrackerButton):
                component.increment(self.increment)
                break
        else:
            # Unreachable, a page tracker should always be sent alongside
            # any page advancing buttons.
            raise RuntimeError

        new_page = await create_paginator(component.page, component.pages)
        await inter.response.edit_message(components=new_page)


@manager.register()
class PageTrackerButton(disnake_compass.RichButton):
    disabled: bool = True

    page: int
    pages: int

    def __attrs_post_init__(self) -> None:
        self.update_label()

    def increment(self, inc: int):
        self.page = (self.page + inc) % self.pages
        self.update_label()

    def update_label(self):
        self.label = f"{self.page + 1} / {self.pages}"

    async def callback(self, _: disnake.MessageInteraction[disnake.Client]) -> None:
        raise NotImplementedError  # This button is always disabled.


def create_paginator_content(page_number: int):
    if page_number == 0:
        return disnake.ui.Section(
            disnake.ui.TextDisplay(
                "Welcome aboard, greenhorn!\nYou are able-bodied and ready for the voyage, I trust?"
            ),
            accessory=disnake.ui.Thumbnail(
                "https://limbuscompany.wiki.gg/images/The_Pequod_Captain_Ishmael_Idle_Sprite.png"
            ),
        )

    if page_number == 1:
        return disnake.ui.Section(
            disnake.ui.TextDisplay("Of course. Has my compass ever led you astray?"),
            accessory=disnake.ui.Thumbnail(
                "https://limbuscompany.wiki.gg/images/The_Pequod_Captain_Ishmael_Evade_Sprite.png"
            ),
        )

    if page_number == 2:
        return disnake.ui.Section(
            disnake.ui.TextDisplay("I'm Fishmael"),
            accessory=disnake.ui.Thumbnail(
                "https://media.tenor.com/mINH1nt-0zgAAAAd/fishmael-limbus-company.gif"
            ),
        )

    msg = f"Unknown page number: {page_number}"
    raise ValueError(msg)


async def create_paginator_controls(page_number: int, pages: int):
    return disnake.ui.ActionRow(
        await PageAdvanceButton(label="<", increment=-1).as_ui_component(),
        await PageTrackerButton(page=page_number, pages=pages).as_ui_component(),
        await PageAdvanceButton(label=">", increment=1).as_ui_component(),
    )


async def create_paginator(page_number: int, pages: int):
    assert page_number < pages

    return disnake.ui.Container(
        disnake.ui.TextDisplay("## Compass"),
        disnake.ui.Separator(),
        create_paginator_content(page_number),
        disnake.ui.Separator(),
        await create_paginator_controls(page_number, pages),
    )


@bot.slash_command()
async def test_v2(inter: disnake.CommandInteraction[disnake.Client]):
    await inter.response.send_message(components=[await create_paginator(0, 3)])


bot.run(os.getenv("EXAMPLE_TOKEN"))



# pyright: reportImportCycles = false
# pyright: reportWildcardImportFromLibrary = false
# ^ This is a false positive as it is confused with site-packages' disnake.

"""Implementations for all kinds of component classes."""

from disnake_compass.impl.component.base import *
from disnake_compass.impl.component.button import *
from disnake_compass.impl.component.select import *



"""Parser implementations for standard library and disnake enums and flags."""

import enum
import typing

import attrs
import disnake

from disnake_compass.api import parser as parser_api
from disnake_compass.impl.parser import base as parser_base

__all__: typing.Sequence[str] = ("EnumParser", "FlagParser")

_AnyEnum: typing.TypeAlias = enum.Enum | disnake.Enum | disnake.flags.BaseFlags
_EnumT = typing.TypeVar("_EnumT", bound=_AnyEnum)


def _get_enum_type(enum_class: type[_AnyEnum]) -> type | None:
    if issubclass(enum_class, disnake.flags.BaseFlags):
        return int

    maybe_type: type = getattr(enum_class, "_member_type_", object)
    if maybe_type is not object:
        return maybe_type

    # Get first member's type
    member_iter = iter(enum_class)
    maybe_type = typing.cast(  # python typing sucks.
        type[typing.Any],
        type(next(member_iter).value),
    )

    # TODO: Check if this can be `is` instead of `==`.
    # If all members match this type, return it.
    if all(type(member.value) == maybe_type for member in member_iter):  # noqa: E721
        return maybe_type

    # No single type; store by name instead.
    return None


@parser_base.register_parser_for(
    enum.Enum,
    disnake.Enum,
    enum.Flag,
    disnake.flags.BaseFlags,
    priority=20,
)
@attrs.define(slots=True, init=False)
class EnumParser(parser_base.Parser[_EnumT]):
    """Parser type for enums and flags.

    Enums and flags are stored by value instead of by name. This makes parsing
    a bit slower, but values are generally shorter than names.

    This parser type works for standard library and disnake enums and flags.
    Note that this only works for enums and flags where all values are of the
    same type.

    Parameters
    ----------
    enum_class:
        The enum or flag class to use for parsing.
    store_by_value:
        Whether :meth:`loads` and :meth:`dumps` expect the enum value type or a string.

        For enum types where the members are *not* all of the same type, this
        *must* be ``False``.

        For enum types where all members are integers, this defaults to
        ``True``, otherwise this defaults to ``False``.

    """

    enum_class: type[_EnumT]
    """The enum or flag class to use for parsing."""
    store_by_value: bool
    """Whether :meth:`loads` and :meth:`dumps` expect the enum's value type or a string.

    For enum types where the members are *not* all of the same type, this
    *must* be ``False``.
    """
    value_parser: parser_api.Parser[typing.Any]
    """The parser responsible for converting to/from the enum type.

    If :attr:`store_by_values` is set to ``False``, this is *always* a
    :class:`~disnake_compass.parser.StringParser`.
    """

    def __init__(
        self,
        enum_class: type[_EnumT],
        *,
        store_by_value: bool | None = None,
    ) -> None:
        if issubclass(enum_class, disnake.flags.BaseFlags) and store_by_value is False:
            msg = "Cannot store disnake flags by name, as their members do not have names."
            raise ValueError(msg)

        value_type = _get_enum_type(enum_class)

        # If the enum is not of a single type and store_by_value wasn't
        # explicitly set or was explicitly set to false, parse by name.
        if value_type is None and not store_by_value:
            self.store_by_value = False
            value_type = str

        # If the value type could be determined, set it on the parser and
        # respect the input for store_by_value-- default to True for integers.
        elif value_type is not None:
            self.store_by_value = (
                issubclass(value_type, int)
                if store_by_value is None
                else store_by_value
            )  # fmt: skip

        # If the value type could not be determined but the user explicitly set
        # store_by_value to True, raise an exception.
        else:
            msg = (
                "Cannot store enum members by value for enums with members of"
                " varying types. Please either set `store_by_value` to False"
                " or make sure all enum members are of the same type."
            )
            raise ValueError(msg)

        self.enum_class = enum_class
        self.value_parser = parser_base.get_parser(value_type)

    async def loads(self, argument: str, /) -> _EnumT:
        """Load an enum member from a string.

        This uses the underlying :attr:`value_parser`.

        .. note::
            If :attr:`store_by_value` is True, the ``argument`` is expected to
            be the value of an enum member; otherwise, the ``argument`` is
            expted to be the name.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a channel.

            This always matches the channel type of the parser.

        """
        parsed = await self.value_parser.loads(argument)

        if self.store_by_value:
            return self.enum_class(parsed)  # pyright: ignore[reportCallIssue]
        return self.enum_class[parsed]  # pyright: ignore[reportInvalidTypeArguments]

    async def dumps(self, argument: _EnumT) -> str:
        """Dump an enum member into a string.

        .. note::
            If :attr:`store_by_value` is True, this dumps the name of the
            enum member passed as ``argument``; otherwise, this dumps its
            value.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        if self.store_by_value:
            return await self.value_parser.dumps(argument.value)
        # Baseflags members are always integers. This should never error
        # due to the check in __init__.
        assert not isinstance(argument, disnake.flags.BaseFlags)
        return await self.value_parser.dumps(argument.name)


FlagParser = EnumParser



"""This module does three "important" things

- Provide custom documenters for enums and flags,
- Change class documenter's ordering to prioritise attributes,
- Add 'section labels' within classes.
"""

import enum
import typing

import sphinx
import sphinx.application
import sphinx.config
import sphinx.domains.python
import sphinx_autodoc_typehints
from sphinx.ext import autodoc
from sphinx.locale import _


class ClassDocumenterWithExtraSteps(autodoc.ClassDocumenter):
    priority = autodoc.ClassDocumenter.priority + 1

    order_display_map: typing.ClassVar[dict[int, str]] = {
        50: _("Methods"),
        60: _("Attributes"),
    }

    def document_members(self, all_members: bool = False) -> None:  # noqa: FBT001, FBT002
        if self.doc_as_attr:
            return

        if (
            self.options.member_order
            or self.config.autodoc_member_order
        ) != "groupwise":  # fmt: skip
            super().document_members(all_members)
            return

        # set current namespace for finding members
        self.env.temp_data["autodoc:module"] = self.modname
        if self.objpath:
            self.env.temp_data["autodoc:class"] = self.objpath[0]

        want_all = (
            all_members or self.options.inherited_members or self.options.members is autodoc.ALL
        )
        # find out which members are documentable
        members_check_module, members = self.get_object_members(want_all)

        # document non-skipped members
        memberdocumenters: list[tuple[autodoc.Documenter, bool]] = []
        for mname, member, isattr in self.filter_members(members, want_all):
            classes = [
                cls
                for cls in self.documenters.values()
                if cls.can_document_member(member, mname, isattr, self)
            ]
            if not classes:
                # don't know how to document this member
                continue
            # prefer the documenter with the highest priority
            # give explicitly separated module name, so that members
            # of inner classes can be documented
            full_mname = f"{self.modname}::" + ".".join((*self.objpath, mname))
            documenter_cls = max(classes, key=lambda cls: cls.priority)
            documenter = documenter_cls(self.directive, full_mname, self.indent)
            memberdocumenters.append((documenter, isattr))

        # NOTE: Personal preference; I'd rather have attributes first.
        memberdocumenters.sort(key=lambda e: (-e[0].member_order, e[0].name))
        source_name = self.get_sourcename()
        last_member_group = 0

        for documenter, isattr in memberdocumenters:
            if documenter.member_order != last_member_group:
                last_member_group = documenter.member_order
                category = self.order_display_map.get(last_member_group)
                if not category:
                    autodoc.logger.warning(
                        "Found member order without display name %i: %s",
                        last_member_group,
                        documenter.objtype,
                    )
                    category = documenter.objtype

                # TODO: Maybe figure out how to actually render this as some
                #       small title thing.
                underline = "^" * len(category)
                self.add_line(f".. _{self.name}-{category}:", source_name)
                self.add_line("", source_name)
                self.add_line(category, source_name)
                self.add_line(underline, source_name)
                self.add_line("", source_name)

            documenter.generate(
                all_members=True,
                real_modname=self.real_modname,
                check_module=members_check_module and not isattr,
            )

        # reset current objects
        self.env.temp_data["autodoc:module"] = None
        self.env.temp_data["autodoc:class"] = None


class EnumDocumenter(ClassDocumenterWithExtraSteps):
    objtype = "enum"
    directivetype = "enum"
    priority = 20
    class_xref = ":class:`~enum.Enum`"

    order_display_map: typing.ClassVar[dict[int, str]] = {
        50: "METHODS",
        60: "MEMBERS",
    }

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        if not self.options.member_order:
            self.options.member_order = "bysource"  # type: ignore

    @classmethod
    def can_document_member(
        cls,
        member: object,
        membername: str,  # noqa: ARG003
        isattr: bool,  # noqa: ARG003, FBT001
        parent: object,  # noqa: ARG003
    ) -> bool:
        return isinstance(member, enum.EnumMeta) and not issubclass(member, enum.Flag)

    def add_content(self, more_content: autodoc.StringList | None) -> None:
        super().add_content(more_content)

        source_name = self.get_sourcename()
        base = self.object.__bases__[0]
        if not issubclass(base, enum.Enum):
            ann = sphinx_autodoc_typehints.format_annotation(base, self.config)
            self.add_line(f"Member type: {ann}", source_name)
            self.add_line("", source_name)


class FlagDocumenter(EnumDocumenter):
    objtype = "flag"
    directivetype = "flag"
    priority = 25
    class_xref = ":class:`~enum.Flag`"

    @classmethod
    def can_document_member(
        cls,
        member: object,
        membername: str,  # noqa: ARG003
        isattr: bool,  # noqa: ARG003, FBT001
        parent: object,  # noqa: ARG003
    ) -> bool:
        return isinstance(member, enum.EnumMeta) and issubclass(member, enum.Flag)


def setup(app: sphinx.application.Sphinx) -> dict[str, bool]:
    app.setup_extension("sphinx.ext.autodoc")
    app.add_autodocumenter(ClassDocumenterWithExtraSteps, override=True)
    app.add_autodocumenter(EnumDocumenter)
    app.add_autodocumenter(FlagDocumenter)

    type_registry = app.registry.domains["py"].object_types
    type_registry["enum"] = sphinx.domains.ObjType(_("enum"), "enum", "class", "obj")
    type_registry["flag"] = sphinx.domains.ObjType(_("flag"), "flag", "class", "obj")

    app.add_directive_to_domain("py", "enum", sphinx.domains.python.PyClasslike)
    app.add_directive_to_domain("py", "flag", sphinx.domains.python.PyClasslike)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""A simple example on the use of buttons with disnake-compass."""

import os

import disnake
import disnake_compass
from disnake.ext import commands

bot = commands.InteractionBot()

manager = disnake_compass.get_manager()
manager.add_to_client(bot)


@manager.register
class MyButton(disnake_compass.RichButton):
    label: str | None = "0"

    count: int = 0

    async def callback(
        self, interaction: disnake.MessageInteraction[disnake.Client]
    ) -> None:
        self.count += 1
        self.label = str(self.count)

        component = await self.as_ui_component()
        await interaction.response.edit_message(components=component)


@bot.slash_command()
async def test_button(interaction: disnake.CommandInteraction[disnake.Client]) -> None:
    component = await MyButton().as_ui_component()
    await interaction.response.send_message(components=component)


bot.run(os.getenv("EXAMPLE_TOKEN"))



"""Lightweight contextvar-based dependency injection-adjacent solution."""

from __future__ import annotations

import contextvars
import typing

import typing_extensions

from disnake_compass.internal import omit

__all__: typing.Sequence[str] = (
    "register_dependencies",
    "reset_dependencies",
    "resolve_dependency",
)

_T = typing.TypeVar("_T")
TokenMap: typing_extensions.TypeAlias = typing.Mapping[
    type[typing.Any],
    contextvars.Token[object],
]

# XXX:  This actually introduces a memory leak when combined with hot-reloading
#       as old types will never be cleared from the dependency map. Since hot-
#       reloading in production environments is ill-advised anyway, I don't
#       think it's worth adding a lot more complexity to circumvent it.
#       ...Unless, of course, someone can come up with a better approach.
DEPENDENCY_MAP: dict[type[typing.Any], contextvars.ContextVar[typing.Any]] = {}


def _get_contextvar_for(dependency_type: type[_T], /) -> contextvars.ContextVar[_T]:
    if dependency_type in DEPENDENCY_MAP:
        return DEPENDENCY_MAP[dependency_type]

    # Resolve subclass of registered type and save it to speed up future lookups.
    for registered_type, context in DEPENDENCY_MAP.items():
        if issubclass(registered_type, dependency_type):
            DEPENDENCY_MAP[dependency_type] = context
            return context

    # Insert missing dependency type, use a name that is unlikely to lead to conflicts.
    name = f"__disnake_ext_components__{dependency_type.__name__}__"
    context = DEPENDENCY_MAP[dependency_type] = contextvars.ContextVar[_T](name)
    return context


def register_dependencies(*dependencies: object) -> TokenMap:
    r"""Register any number of dependencies.

    This returns a mapping of :class:`contextvars.Token`\s that should be
    passed to :func:`reset_dependencies` for cleanup.

    While dependencies are registered, :func:`resolve_dependency` can be used
    to get it for the current async context.

    Parameters
    ----------
    *dependencies:
        Objects to register as dependencies. Their type must be hashable.
        (This should automatically hold for most classes.)

    Returns
    -------
    typing.Dict[typing.Type[typing.Any], contextvars.Token[object]]
        A mapping of the types of the registered dependencies to contextvar
        tokens used to reset their respective contextvars. This is meant to be
        passed to :func:`reset_dependencies` for cleanup.

    """
    tokens: dict[type, contextvars.Token[object]] = {}
    for dependency in dependencies:
        dependency_type = type(dependency)
        tokens[dependency_type] = _get_contextvar_for(dependency_type).set(dependency)

    return tokens


def reset_dependencies(tokens: TokenMap) -> None:
    """Reset dependencies that are no longer in use.

    This is meant to be used in conjunction with :func:`register_dependencies`.

    Parameters
    ----------
    tokens:
        A mapping of the types of the registered dependencies to contextvar
        tokens used to reset their respective contextvars. This mapping is
        created and returned by :func:`register_dependencies`.

    """
    for dependency_type, token in tokens.items():
        _get_contextvar_for(dependency_type).reset(token)


def resolve_dependency(
    dependency_type: type[_T],
    default: omit.Omissible[_T] = omit.Omitted,
) -> _T:
    """Resolve a dependency given a type and an optional default.

    If a dependency was set using :func:`register_dependency` in the current
    context, this function returns it. If it is not found, the default is
    returned instead. If no default was provided, a :class:`LookupError` is
    raised instead.

    Parameters
    ----------
    dependency_type:
        The type to resolve to an object.
    default:
        The default to use if resolving did not return an object.

    Returns
    -------
    object
        The resolved dependency or the default.

    Raises
    ------
    :class:`LookupError`
        The dependency type could not be resolved and no default was provided.

    """
    context = _get_contextvar_for(dependency_type)
    resolved = context.get(omit.Omitted)
    if not omit.is_omitted(resolved):
        return resolved

    if not omit.is_omitted(default):
        return default

    msg = f"Failed to resolve dependency for type {dependency_type.__name__}."
    raise LookupError(msg)



# pyright: reportImportCycles = false
# pyright: reportWildcardImportFromLibrary = false
# ^ This is a false positive as it is confused with site-packages' disnake.

"""Default concrete implementations for types in ``disnake_compass.api``."""

from disnake_compass.impl import parser as parser
from disnake_compass.impl.component import *
from disnake_compass.impl.factory import *
from disnake_compass.impl.manager import *
